"""
Helpers for the asset mutations
"""
import csv
import mimetypes
import os
from json import dumps
from typing import List, Union
from uuid import uuid4

from ...helpers import (
    check_file_mime_type,
    convert_to_list_of_none,
    encode_base64,
    format_metadata,
    is_none_or_empty,
    is_url,
)
from .queries import GQL_APPEND_MANY_FRAMES_TO_DATASET, GQL_APPEND_MANY_TO_DATASET


def encode_object_if_not_url(content, input_type):
    """
    Return the object if it is a url, else it should be a path to a file.
    In that case, the file is returned as a base64 string
    """
    if is_url(content):
        return content
    if check_file_mime_type(content, input_type):
        return encode_base64(content)
    return None


def process_frame_json_content(json_content):
    """
    Function to process individual json_content of VIDEO projects
    """
    if is_url(json_content):
        return json_content
    json_content_index = range(len(json_content))
    json_content_urls = [encode_object_if_not_url(content, "IMAGE") for content in json_content]
    return dumps(dict(zip(json_content_index, json_content_urls)))


def get_file_mimetype(
    content_array: Union[List[str], None], json_content_array: Union[List[str], None]
) -> Union[str, None]:
    """
    Returns the mimetype of the first file of the content array
    """
    if json_content_array is not None:
        return None
    if content_array is None:
        return None
    if len(content_array) > 0:
        first_asset = content_array[0]
        if is_url(first_asset):
            return None
        if not os.path.exists(first_asset):
            return None
        return mimetypes.guess_type(first_asset)[0]
    return None


def process_json_content(
    input_type: str,
    content_array: List[str],
    json_content_array: Union[List[str], None],
):
    """
    Process the array of json_contents
    """
    if json_content_array is None:
        return [""] * len(content_array)
    if input_type in ("FRAME", "VIDEO"):
        return list(map(process_frame_json_content, json_content_array))
    return [element if is_url(element) else dumps(element) for element in json_content_array]


def process_content(
    input_type: str,
    content_array: Union[List[str], None],
    json_content_array: Union[List[List[Union[dict, str]]], None],
):
    """
    Process the array of contents
    """
    if input_type in ["IMAGE", "PDF"]:
        return [
            content
            if is_url(content)
            else (
                content
                if (json_content_array is not None and json_content_array[i] is not None)
                else (encode_base64(content) if check_file_mime_type(content, input_type) else None)
            )
            for i, content in enumerate(content_array)
        ]
    if input_type in ("FRAME", "VIDEO") and json_content_array is None:
        content_array = [encode_object_if_not_url(content, input_type) for content in content_array]
    if input_type == "TIME_SERIES":
        content_array = list(map(process_time_series, content_array))
    return content_array


def process_time_series(content: str) -> Union[str, None]:
    """
    Process the content for TIME_SERIES projects: if it is a file, read the content
    and also check if the content corresponds to the expected format, else return None
    """
    delimiter = ","
    if os.path.isfile(content):
        if check_file_mime_type(content, "TIME_SERIES"):
            with open(content, "r", encoding="utf8") as csvfile:
                reader = csv.reader(csvfile, delimiter=delimiter)
                return process_csv_content(reader, file_name=content, delimiter=delimiter)
        return None

    reader = csv.reader(content.split("\n"), delimiter=",")
    return process_csv_content(reader, delimiter=delimiter)


def process_csv_content(reader, file_name=None, delimiter=",") -> bool:
    """
    Process the content of csv for time_series and check if it corresponds to the expected format
    """
    first_row = True
    processed_lines = []
    for row in reader:
        if not (len(row) == 2 and (first_row or (not first_row and is_float(row[0])))):
            print(
                f"""The content {file_name if file_name else row} does not correspond to the \
correct format: it should have only 2 columns, the first one being the timestamp \
(an integer or a float) and the second one a numeric value (an integer or a float, \
otherwise it will be considered as missing value). The first row should have the names \
of the 2 columns. The delimiter used should be ','."""
            )
            return None
        value = row[1] if (is_float(row[1]) or first_row) else ""
        processed_lines.append(delimiter.join([row[0], value]))
        first_row = False
    return "\n".join(processed_lines)


def is_float(number: str) -> bool:
    """
    Check if a string can be converted to float
    """
    try:
        float(number)
        return True
    except ValueError:
        return False


def add_video_parameters(json_metadata, should_use_native_video):
    """
    Add necessary video parameters to the metadata of the video
    """
    processing_parameters = json_metadata.get("processingParameters", {})
    video_parameters = [
        ("shouldKeepNativeFrameRate", should_use_native_video),
        ("framesPlayedPerSecond", 30),
        ("shouldUseNativeVideo", should_use_native_video),
    ]
    for (key, default_value) in video_parameters:
        processing_parameters[key] = processing_parameters.get(key, default_value)
    return {**json_metadata, "processingParameters": processing_parameters}


def process_metadata(
    input_type: str,
    content_array: Union[List[str], None],
    json_content_array: Union[List[List[Union[dict, str]]], None],
    json_metadata_array: Union[List[dict], None],
):
    """
    Process the metadata of each asset
    """
    json_metadata_array = (
        [{}] * len(content_array) if json_metadata_array is None else json_metadata_array
    )
    if input_type in ("FRAME", "VIDEO"):
        should_use_native_video = json_content_array is None
        json_metadata_array = [
            add_video_parameters(json_metadata, should_use_native_video)
            for json_metadata in json_metadata_array
        ]
    return list(map(dumps, json_metadata_array))


def get_request_to_execute(
    input_type: str,
    json_metadata_array: Union[List[dict], None],
    json_content_array: Union[List[List[Union[dict, str]]], None],
    mime_type: Union[str, None],
) -> str:
    """
    Selects the right query to run versus the data given
    """
    if json_content_array is not None:
        return GQL_APPEND_MANY_TO_DATASET, None
    if input_type not in ("FRAME", "VIDEO"):
        if input_type == "IMAGE" and mime_type == "image/tiff":
            return GQL_APPEND_MANY_FRAMES_TO_DATASET, "GEO_SATELLITE"
        return GQL_APPEND_MANY_TO_DATASET, None
    if json_metadata_array is None:
        return GQL_APPEND_MANY_TO_DATASET, None
    if (
        isinstance(json_metadata_array, list)
        and len(json_metadata_array) > 0
        and not json_metadata_array[0]
        .get("processingParameters", {})
        .get("shouldUseNativeVideo", True)
    ):
        return GQL_APPEND_MANY_FRAMES_TO_DATASET, "VIDEO"
    return GQL_APPEND_MANY_TO_DATASET, None


# pylint: disable=too-many-arguments
def process_append_many_to_dataset_parameters(
    input_type: str,
    content_array: Union[List[str], None],
    external_id_array: Union[List[str], None],
    is_honeypot_array: Union[List[str], None],
    status_array: Union[List[str], None],
    json_content_array: Union[List[List[Union[dict, str]]], None],
    json_metadata_array: Union[List[dict], None],
):
    """
    Process arguments of the append_many_to_dataset method and return the data payload.
    """
    if content_array is None and json_content_array is None:
        raise ValueError("Variables content_array and json_content_array cannot be both None.")
    if content_array is None:
        content_array = [""] * len(json_content_array)
    if external_id_array is None:
        external_id_array = [uuid4().hex for _ in range(len(content_array))]
    is_honeypot_array = (
        [False] * len(content_array) if is_honeypot_array is None else is_honeypot_array
    )
    status_array = ["TODO"] * len(content_array) if not status_array else status_array
    formatted_json_metadata_array = process_metadata(
        input_type, content_array, json_content_array, json_metadata_array
    )
    mime_type = get_file_mimetype(content_array, json_content_array)
    content_array = process_content(input_type, content_array, json_content_array)
    formatted_json_content_array = process_json_content(
        input_type, content_array, json_content_array
    )

    request, upload_type = get_request_to_execute(
        input_type, json_metadata_array, json_content_array, mime_type
    )
    properties = {
        "content_array": content_array,
        "external_id_array": external_id_array,
        "is_honeypot_array": is_honeypot_array,
        "status_array": status_array,
        "json_content_array": formatted_json_content_array,
        "json_metadata_array": formatted_json_metadata_array,
    }

    return properties, upload_type, request


def process_update_properties_in_assets_parameters(properties) -> dict:
    """
    Process arguments of the update_properties_in_assets method
    and return the properties for the paginated loop
    """
    formatted_json_metadatas = None
    if properties["json_metadatas"] is None:
        formatted_json_metadatas = None
    else:
        if isinstance(properties["json_metadatas"], list):
            formatted_json_metadatas = list(map(format_metadata, properties["json_metadatas"]))
        else:
            raise Exception(
                "json_metadatas",
                "Should be either a None or a list of None, string, list or dict",
            )
    properties["json_metadatas"] = formatted_json_metadatas
    nb_assets_to_modify = len(properties["asset_ids"])
    properties = {
        k: convert_to_list_of_none(v, length=nb_assets_to_modify) for k, v in properties.items()
    }
    properties["should_reset_to_be_labeled_by_array"] = list(
        map(is_none_or_empty, properties["to_be_labeled_by_array"])
    )
    return properties


def generate_json_metadata_array(as_frames, fps, nb_files, input_type):
    """Generate the json_metadata_array for input of the append_many_to_dataset resolver
    when uploading from a list of path

    Args:
        as_frames: for a frame project, if videos should be split in frames
        fps: for a frame project, import videos with this frame rate
        nb_files: the number of files to upload in the call
        input_type: the input type of the project to upload to
    """

    json_metadata_array = None
    if input_type in ("FRAME", "VIDEO"):
        json_metadata_array = [
            {
                "processingParameters": {
                    "shouldKeepNativeFrameRate": fps is None,
                    "framesPlayedPerSecond": fps,
                    "shouldUseNativeVideo": not as_frames,
                }
            }
        ] * nb_files
    return json_metadata_array



"""
Test notebooks with pytest
"""

import os
import subprocess

import nbformat
from nbconvert.preprocessors import CellExecutionError, ExecutePreprocessor

RECIPES_DIR = "./recipes"
RECIPES_TESTED = [
    "create_project.ipynb",
    "export_labels.ipynb",
    "frame_dicom_data.ipynb",
    "import_assets.ipynb",
    "import_predictions.ipynb",
    "import_text_assets.ipynb",
    "inference_labels.ipynb",
    "medical_imaging.ipynb",
    "ocr_pre_annotations.ipynb",
    "pixel_level_masks.ipynb",
    "query_methods.ipynb",
    "webhooks.ipynb",
    "getting-started/getting_started-image_bounding_box_detection.ipynb",
    "getting-started/getting_started-image_classification.ipynb",
    "getting-started/getting_started-image_semantic_segmentation.ipynb",
    "getting-started/getting_started-named-entity-recognition.ipynb",
]


def process_notebook(notebook_filename):
    """
    Checks if an IPython notebook runs without error from start to finish.
    """
    with open(notebook_filename) as f:
        nb = nbformat.read(f, as_version=4)

    ep = ExecutePreprocessor(timeout=1000, kernel_name="python3")

    try:
        # Check that the notebook runs
        ep.preprocess(nb, {"metadata": {"path": ""}})
    except CellExecutionError:
        raise

    print(f"Successfully executed {notebook_filename}")
    return


def test_all_recipes():
    """
    Runs `process_notebook` on all notebooks in the git repository.
    """
    notebooks = [os.path.join(RECIPES_DIR, recipe) for recipe in RECIPES_TESTED]
    for notebook in notebooks:
        print("Testing", notebook)
        process_notebook(notebook)

    return


if __name__ == "__main__":
    test_all_recipes()



"""
Lock queries
"""

from typing import Generator, List, Optional, Union

from typeguard import typechecked

from ...helpers import Compatible, format_result, fragment_builder
from ...types import Lock
from ...utils.pagination import row_generator_from_paginated_calls
from .queries import GQL_LOCKS_COUNT, gql_locks


class QueriesLock:
    """
    Set of Lock queries
    """

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(["v1", "v2"])
    @typechecked
    def locks(
        self,
        lock_id: Optional[str] = None,
        fields: List[str] = ["id", "lockType"],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of locks respecting a set of criteria.

        Args:
            lock_id: The id of the lock to request. If None, all locks are returned
            fields: All the fields to request among the possible fields for the locks.
                See [the documentation](https://cloud.kili-technology.com/docs/python-graphql-api/graphql-api/#locks)
                for all possible fields.
            first: Maximum number of locks to return.
            skip: Number of skipped locks (they are ordered by creation date)
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the API key is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.
        """

        count_args = {}
        payload_query = {"where": {"id": lock_id}}

        disable_tqdm = disable_tqdm or as_generator or lock_id is not None

        locks_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_locks,
            count_args,
            self._query_locks,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return locks_generator
        return list(locks_generator)

    def _query_locks(self, skip: int, first: int, payload: dict, fields: List[str]):

        payload.update({"skip": skip, "first": first})
        _gql_locks = gql_locks(fragment_builder(fields, Lock))
        result = self.auth.client.execute(_gql_locks, payload)
        return format_result("data", result)

    @Compatible(["v1", "v2"])
    @typechecked
    def count_locks(self: any) -> int:
        """Get the number of locks

        Args:

        Returns:
            The number of locks
        """
        variables = {"where": {"id": None}}
        result = self.auth.client.execute(GQL_LOCKS_COUNT, variables)
        count = format_result("data", result)
        return count



"""
Queries of apiKey mutations
"""

from .fragments import API_KEY_FRAGMENT

GQL_APPEND_TO_API_KEYS = f"""
mutation(
  $data: AppendToApiKeysData!
  $where: UserWhere!
) {{
  data: appendToApiKeys(data: $data, where: $where) {{
    {API_KEY_FRAGMENT}
  }}
}}
"""



"""Tests the Kili CLI"""

import csv
import json
import os
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from kili.cli.project.create import create_project
from kili.cli.project.describe import describe_project
from kili.cli.project.import_ import import_assets
from kili.cli.project.label import import_labels
from kili.cli.project.list_ import list_projects

from ..utils import debug_subprocess_pytest

mock_label = {"JOB_0": {"categories": [{"name": "YES_IT_IS_SPAM", "confidence": 100}]}}


def mocked__projects(project_id=None, **_):
    if project_id == "text_project":
        return [{"id": "text_project", "inputType": "TEXT"}]
    if project_id == "image_project":
        return [{"id": "image_project", "inputType": "IMAGE"}]
    if project_id == "frame_project":
        return [{"id": "frame_project", "inputType": "VIDEO"}]
    if project_id == None:
        return [
            {
                "id": "text_project",
                "title": "text_project",
                "description": " a project with text",
                "numberOfAssets": 10,
                "numberOfRemainingAssets": 10,
            },
            {
                "id": "image_project",
                "title": "image_project",
                "description": " a project with image",
                "numberOfAssets": 0,
                "numberOfRemainingAssets": 0,
            },
            {
                "id": "frame_project",
                "title": "frame_project",
                "description": " a project with frame",
                "numberOfAssets": 10,
                "numberOfRemainingAssets": 0,
            },
        ]
    if project_id == "project_id":
        return [
            {
                "title": "project title",
                "id": "project_id",
                "description": "description test",
                "numberOfAssets": 49,
                "numberOfRemainingAssets": 29,
                "numberOfReviewedAssets": 0,
                "numberOfAssetsWithSkippedLabels": 0,
                "numberOfOpenIssues": 3,
                "numberOfSolvedIssues": 2,
                "numberOfOpenQuestions": 0,
                "numberOfSolvedQuestions": 2,
                "honeypotMark": None,
                "consensusMark": None,
            }
        ]


def mocked__project_assets(**_):
    return [
        {"externalId": "asset1"},
        {"externalId": "asset2"},
        {"externalId": "asset3"},
        {"externalId": "asset4"},
        {"externalId": "asset5"},
        {"externalId": "asset6"},
    ]


kili_client = MagicMock()
kili_client.auth.client.endpoint = "https://staging.cloud.kili-technology.com/api/label/v2/graphql"
kili_client.projects = project_mock = MagicMock(side_effect=mocked__projects)
kili_client.append_to_labels = append_to_labels_mock = MagicMock()
kili_client.create_predictions = create_predictions_mock = MagicMock()
kili_client.count_projects = count_projects_mock = MagicMock(return_value=1)
kili_client.append_many_to_dataset = append_many_to_dataset_mock = MagicMock()
kili_client.create_project = create_project_mock = MagicMock()
kili_client.assets = assets_mock = MagicMock(side_effect=mocked__project_assets)

kili = MagicMock()
kili.Kili = MagicMock(return_value=kili_client)


@patch("kili.client.Kili.__new__", return_value=kili_client)
class TestCLIProject:
    """
    test the CLI functions of the project command
    """

    def test_list(self, mocker):
        runner = CliRunner()
        result = runner.invoke(list_projects)
        assert (
            (result.exit_code == 0)
            and (result.output.count("100.0%") == 1)
            and (result.output.count("0.0%") == 2)
            and (result.output.count("nan") == 1)
        )

    def test_create_project(self, mocker):
        runner = CliRunner()
        runner.invoke(
            create_project,
            [
                "--interface",
                "test/fixtures/image_interface.json",
                "--title",
                "Test project",
                "--description",
                "description",
                "--input-type",
                "IMAGE",
            ],
        )

    def test_describe_project(self, mocker):
        runner = CliRunner()
        result = runner.invoke(describe_project, ["project_id"])
        debug_subprocess_pytest(result)
        assert (
            (result.output.count("40.8%") == 1)
            and (result.output.count("N/A") == 2)
            and (result.output.count("49") == 1)
            and (result.output.count("project title") == 1)
        )

    def test_import_labels(self, mocker):
        TEST_CASES = [
            {
                "case_name": "AAU, when I import a list of label's file to project, I see a success",
                "files": ["test_tree/leaf_1/", "test_tree/asset1.json", "test_tree/leaf_2/**.json"],
                "options": {
                    "project-id": "project_id",
                },
                "flags": [],
                "mutation_to_call": "append_to_labels",
                "expected_mutation_payload": {
                    "project_id": "project_id",
                    "json_response": {
                        "JOB_0": {"categories": [{"name": "YES_IT_IS_SPAM", "confidence": 100}]}
                    },
                    "label_asset_external_id": "asset6",
                },
            },
            {
                "case_name": "AAU, when I import default labels from a CSV, I see a success",
                "files": [],
                "options": {
                    "project-id": "project_id",
                    "from-csv": "labels_to_import.csv",
                },
                "flags": [],
                "mutation_to_call": "append_to_labels",
                "expected_mutation_payload": {
                    "project_id": "project_id",
                    "json_response": {
                        "JOB_0": {"categories": [{"name": "YES_IT_IS_SPAM", "confidence": 100}]}
                    },
                    "label_asset_external_id": "asset6",
                },
            },
            {
                "case_name": "AAU, when I import predictions from a CSV, I see a sucess",
                "files": [],
                "options": {
                    "project-id": "project_id",
                    "model-name": "model_name",
                    "from-csv": "labels_to_import.csv",
                },
                "flags": ["prediction"],
                "mutation_to_call": "create_predictions",
                "expected_mutation_payload": {
                    "project_id": "project_id",
                    "json_response_array": [
                        {"JOB_0": {"categories": [{"name": "YES_IT_IS_SPAM", "confidence": 100}]}}
                    ]
                    * 2,
                    "external_id_array": ["asset1", "asset6"],
                    "model_name_array": ["model_name"] * 2,
                },
            },
        ]
        runner = CliRunner()
        with runner.isolated_filesystem():
            os.mkdir("test_tree")
            # pylint: disable=unspecified-encoding
            with open("test_tree/asset1.json", "w") as outfile:
                json.dump(mock_label, outfile)
            with open("test_tree/asset2.json", "w") as outfile:
                json.dump(mock_label, outfile)
            os.mkdir("test_tree/leaf_1")
            with open("test_tree/leaf_1/asset3.json", "w") as outfile:
                json.dump(mock_label, outfile)
            with open("test_tree/leaf_1/asset4.json", "w") as outfile:
                json.dump(mock_label, outfile)
            os.mkdir("test_tree/leaf_2")
            with open("test_tree/leaf_2/asset5.json", "w") as outfile:
                json.dump(mock_label, outfile)
            with open("test_tree/leaf_2/asset6.json", "w") as outfile:
                json.dump(mock_label, outfile)
            with open("labels_to_import.csv", "w") as f:
                writer = csv.writer(f)
                writer.writerow(["external_id", "json_response_path"])
                writer.writerow(["asset1", "test_tree/asset1.json"])
                writer.writerow(["asset6", "test_tree/leaf_2/asset6.json"])

            for i, test_case in enumerate(TEST_CASES):
                arguments = test_case["files"]
                for k, v in test_case["options"].items():
                    arguments.append("--" + k)
                    arguments.append(v)
                if test_case.get("flags"):
                    arguments.extend(["--" + flag for flag in test_case["flags"]])
                print(arguments)
                result = runner.invoke(import_labels, arguments)
                debug_subprocess_pytest(result)
                if test_case["mutation_to_call"] == "append_to_labels":
                    assert append_to_labels_mock.call_count == 5 + 2 * i
                    append_to_labels_mock.assert_any_call(**test_case["expected_mutation_payload"])
                else:
                    create_predictions_mock.assert_called_with(
                        **test_case["expected_mutation_payload"]
                    )

    def test_import(self, mocker):
        TEST_CASES = [
            {
                "case_name": "AAU, when I import a list of file to an image project, I see a success",
                "files": ["test_tree/image1.png", "test_tree/leaf/image3.png"],
                "options": {
                    "project-id": "image_project",
                },
                "expected_mutation_payload": {
                    "project_id": "image_project",
                    "content_array": [
                        "test_tree/image1.png",
                        "test_tree/leaf/image3.png",
                    ],
                    "external_id_array": ["image1", "image3"],
                    "json_metadata_array": None,
                },
            },
            {
                "case_name": "AAU, when I import files with stars, I see a success",
                "files": ["test_tree/**.jpg", "test_tree/leaf/**.jpg"],
                "options": {
                    "project-id": "image_project",
                },
                "expected_mutation_payload": {
                    "project_id": "image_project",
                    "content_array": [
                        "test_tree/image2.jpg",
                        "test_tree/leaf/image4.jpg",
                    ],
                    "external_id_array": ["image2", "image4"],
                    "json_metadata_array": None,
                },
            },
            {
                "case_name": "AAU, when I import a files to a text project, I see a success",
                "files": ["test_tree/", "test_tree/leaf"],
                "options": {"project-id": "text_project"},
                "expected_mutation_payload": {
                    "project_id": "text_project",
                    "content_array": [
                        "test_tree/leaf/texte2.txt",
                        "test_tree/texte1.txt",
                    ],
                    "external_id_array": ["texte2", "texte1"],
                    "json_metadata_array": None,
                },
            },
            {
                "case_name": "AAU, when I import videos to a video project, as native by changing the fps, I see a success",
                "files": ["test_tree/"],
                "options": {
                    "project-id": "frame_project",
                    "fps": "10",
                },
                "expected_mutation_payload": {
                    "project_id": "frame_project",
                    "content_array": ["test_tree/video1.mp4", "test_tree/video2.mp4"],
                    "external_id_array": ["video1", "video2"],
                    "json_metadata_array": [
                        {
                            "processingParameters": {
                                "shouldKeepNativeFrameRate": False,
                                "framesPlayedPerSecond": 10,
                                "shouldUseNativeVideo": True,
                            }
                        }
                    ]
                    * 2,
                },
            },
            {
                "case_name": "AAU, when I import videos to a video project, as frames with the native frame rate, I see a success",
                "files": ["test_tree/"],
                "options": {
                    "project-id": "frame_project",
                },
                "flags": ["frames"],
                "expected_mutation_payload": {
                    "project_id": "frame_project",
                    "content_array": ["test_tree/video1.mp4", "test_tree/video2.mp4"],
                    "external_id_array": ["video1", "video2"],
                    "json_metadata_array": [
                        {
                            "processingParameters": {
                                "shouldKeepNativeFrameRate": True,
                                "framesPlayedPerSecond": None,
                                "shouldUseNativeVideo": False,
                            }
                        }
                    ]
                    * 2,
                },
            },
            {
                "case_name": "AAU, when I import assets from a csv file, I see a success",
                "files": [],
                "options": {
                    "project-id": "image_project",
                    "from-csv": "assets_to_import.csv",
                },
                "expected_mutation_payload": {
                    "project_id": "image_project",
                    "content_array": (
                        [
                            "test_tree/leaf/image3.png",
                            "https://files.readme.io/cac9114-Kili_Wordmark_SoftWhite_RGB.svg",
                        ]
                    ),
                    "external_id_array": ["image3", "test"],
                    "json_metadata_array": None,
                },
            },
        ]
        runner = CliRunner()
        with runner.isolated_filesystem():
            os.mkdir("test_tree")
            # pylint: disable=unspecified-encoding
            open("test_tree/image1.png", "w")
            open("test_tree/image2.jpg", "w")
            open("test_tree/texte1.txt", "w")
            open("test_tree/video1.mp4", "w")
            open("test_tree/video2.mp4", "w")
            os.mkdir("test_tree/leaf")
            open("test_tree/leaf/image3.png", "w")
            open("test_tree/leaf/image4.jpg", "w")
            open("test_tree/leaf/texte2.txt", "w")
            with open("assets_to_import.csv", "w") as f:
                writer = csv.writer(f)
                writer.writerow(["external_id", "content"])
                writer.writerow(["image3", "test_tree/leaf/image3.png"])
                writer.writerow(
                    [
                        "test",
                        "https://files.readme.io/cac9114-Kili_Wordmark_SoftWhite_RGB.svg",
                    ]
                )

            for test_case in TEST_CASES:
                arguments = test_case["files"]
                for k, v in test_case["options"].items():
                    arguments.append("--" + k)
                    arguments.append(v)
                if test_case.get("flags"):
                    arguments.extend(["--" + flag for flag in test_case["flags"]])
                print(arguments)
                result = runner.invoke(import_assets, arguments)
                debug_subprocess_pytest(result)
                append_many_to_dataset_mock.assert_called_with(
                    **test_case["expected_mutation_payload"]
                )



"""Project queries."""

from typing import Generator, List, Optional, Union

from typeguard import typechecked

from ...helpers import Compatible, format_result, fragment_builder
from ...types import Project
from ...utils.pagination import row_generator_from_paginated_calls
from .queries import GQL_PROJECTS_COUNT, gql_projects


class QueriesProject:
    """Set of Project queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(["v1", "v2"])
    @typechecked
    def projects(
        self,
        project_id: Optional[str] = None,
        search_query: Optional[str] = None,
        should_relaunch_kpi_computation: Optional[bool] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
        skip: int = 0,
        fields: List[str] = [
            "consensusTotCoverage",
            "id",
            "inputType",
            "jsonInterface",
            "minConsensusSize",
            "reviewCoverage",
            "roles.id",
            "roles.role",
            "roles.user.email",
            "roles.user.id",
            "title",
        ],
        first: Optional[int] = None,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of projects that match a set of criteria.

        Args:
            project_id: Select a specific project through its project_id.
            search_query: Returned projects with a title or a description matching this string.
            should_relaunch_kpi_computation : Technical field, added to indicate changes in honeypot or consensus settings.
            updated_at_gte: Returned projects should have a label whose update date is greater or equal
                to this date.
            updated_at_lte: Returned projects should have a label whose update date is lower or equal to this date.
            skip: Number of projects to skip (they are ordered by their creation).
            fields: All the fields to request among the possible fields for the projects.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#project) for all possible fields.
            first: Maximum number of projects to return.
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the projects is returned.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            >>> # List all my projects
            >>> kili.projects()
        """

        saved_args = locals()
        count_args = {
            k: v
            for (k, v) in saved_args.items()
            if k
            in [
                "project_id",
                "search_query",
                "should_relaunch_kpi_computation",
                "updated_at_gte",
                "updated_at_lte",
            ]
        }
        disable_tqdm = disable_tqdm or as_generator

        payload_query = {
            "where": {
                "id": project_id,
                "searchQuery": search_query,
                "shouldRelaunchKpiComputation": should_relaunch_kpi_computation,
                "updatedAtGte": updated_at_gte,
                "updatedAtLte": updated_at_lte,
            },
        }

        projects_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_projects,
            count_args,
            self._query_projects,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return projects_generator
        return list(projects_generator)

    def _query_projects(self, skip: int, first: int, payload: dict, fields: List[str]):

        payload.update({"skip": skip, "first": first})
        _gql_projects = gql_projects(fragment_builder(fields, Project))
        result = self.auth.client.execute(_gql_projects, payload)
        return format_result("data", result)

    @Compatible(["v1", "v2"])
    @typechecked
    def count_projects(
        self,
        project_id: Optional[str] = None,
        search_query: Optional[str] = None,
        should_relaunch_kpi_computation: Optional[bool] = None,
        updated_at_gte: Optional[str] = None,
        updated_at_lte: Optional[str] = None,
    ) -> int:
        """
        Counts the number of projects with a search_query

        Args:
            project_id: Select a specific project through its project_id.
            search_query: Returned projects have a title or a description that matches this string.
            should_relaunch_kpi_computation : Technical field, added to indicate changes in honeypot
                or consensus settings
            updated_at_gte: Returned projects should have a label
                whose update date is greater
                or equal to this date.
            updated_at_lte: Returned projects should have a label
                whose update date is lower or equal to this date.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            The number of projects with the parameters provided
        """
        variables = {
            "where": {
                "id": project_id,
                "searchQuery": search_query,
                "shouldRelaunchKpiComputation": should_relaunch_kpi_computation,
                "updatedAtGte": updated_at_gte,
                "updatedAtLte": updated_at_lte,
            }
        }
        result = self.auth.client.execute(GQL_PROJECTS_COUNT, variables)
        count = format_result("data", result)
        return count



"""Tests the Kili CLI project member commands"""

import csv
import os
from unittest.mock import MagicMock, patch

from click.testing import CliRunner

from kili.cli.project.member.add import add_member
from kili.cli.project.member.list_ import list_members
from kili.cli.project.member.remove import remove_member
from kili.cli.project.member.update import update_member

from ..utils import debug_subprocess_pytest


def mocked__project_users(project_id=None, **_):
    if project_id == "project_id_source":
        return [
            {
                "activated": True,
                "id": "role_id_src_alice",
                "role": "ADMIN",
                "user": {
                    "email": "alice@test.com",
                    "firstname": "alice",
                    "id": "user_id_alice",
                    "lastname": "alice",
                    "organization": {"name": "test"},
                },
            },
            {
                "activated": True,
                "id": "role_id_src_bob",
                "role": "LABELER",
                "user": {
                    "email": "bob@test.com",
                    "firstname": "bob",
                    "id": "user_id_bob",
                    "lastname": "bob",
                    "organization": {"name": "test"},
                },
            },
            {
                "activated": True,
                "id": "role_id_src_john",
                "role": "REVIEWER",
                "user": {
                    "email": "john.doe@test.com",
                    "firstname": "john",
                    "id": "user_id_john",
                    "lastname": "doe",
                    "organization": {"name": "test"},
                },
            },
            {
                "activated": True,
                "id": "role_id_src_jane",
                "role": "REVIEWER",
                "user": {
                    "email": "jane.doe@test.com",
                    "firstname": "jane",
                    "id": "user_id_jane",
                    "lastname": "doe",
                    "organization": {"name": "test"},
                },
            },
        ]
    elif project_id == "new_project":
        return []
    else:
        return [
            {
                "activated": True,
                "id": "role_id_john",
                "role": "ADMIN",
                "user": {
                    "email": "john.doe@test.com",
                    "firstname": "john",
                    "id": "user_id_john",
                    "lastname": "doe",
                    "organization": {"name": "test"},
                },
            },
            {
                "activated": True,
                "id": "role_id_jane",
                "role": "LABELER",
                "user": {
                    "email": "jane.doe@test.com",
                    "firstname": "jane",
                    "id": "user_id_jane",
                    "lastname": "doe",
                    "organization": {"name": "test"},
                },
            },
        ]


kili_client = MagicMock()
kili_client.auth.client.endpoint = "https://staging.cloud.kili-technology.com/api/label/v2/graphql"
kili_client.project_users = project_users_mock = MagicMock(side_effect=mocked__project_users)
kili_client.append_to_roles = append_to_roles_mock = MagicMock()
kili_client.update_properties_in_role = update_properties_in_role_mock = MagicMock()
kili_client.delete_from_roles = delete_from_roles_mock = MagicMock()

kili = MagicMock()
kili.Kili = MagicMock(return_value=kili_client)


@patch("kili.client.Kili.__new__", return_value=kili_client)
class TestCLIProjectMember:
    """
    test the CLI functions of the project member commands
    """

    def test_list_members(self, mocker):
        runner = CliRunner()
        result = runner.invoke(list_members, ["project_id"])
        debug_subprocess_pytest(result)
        assert (result.output.count("Jane Doe") == 1) and (result.output.count("@test.com") == 2)

    def test_add_member(self, mocker):
        TEST_CASES = [
            {
                "case_name": "AAU, when I add one user with email adress, I see a success",
                "inputs": ["bob@test.com", "alice@test.com"],
                "options": {
                    "project-id": "new_project",
                },
                "expected_mutation_payload": {
                    "project_id": "new_project",
                    "user_email": "alice@test.com",
                    "role": "LABELER",
                },
            },
            {
                "case_name": "AAU, when I add user from another project, I see a success",
                "inputs": [],
                "options": {"project-id": "new_project", "from-project": "project_id_"},
                "expected_mutation_payload": {
                    "project_id": "new_project",
                    "user_email": "jane.doe@test.com",
                    "role": "LABELER",
                },
            },
            {
                "case_name": "AAU, when I add one user with csv file, I see a success",
                "inputs": [],
                "options": {
                    "project-id": "new_project",
                    "from-csv": "user_list.csv",
                    "role": "REVIEWER",
                },
                "expected_mutation_payload": {
                    "project_id": "new_project",
                    "user_email": "bob@test.com",
                    "role": "REVIEWER",
                },
            },
        ]
        runner = CliRunner()
        with runner.isolated_filesystem():
            # pylint: disable=unspecified-encodind
            with open("user_list.csv", "w") as f:
                writer = csv.writer(f)
                writer.writerow(["email"])
                writer.writerow(["alice@test.com"])
                writer.writerow(["bob@test.com"])

            for i, test_case in enumerate(TEST_CASES):
                print(test_case["case_name"])
                arguments = test_case["inputs"]
                for k, v in test_case["options"].items():
                    arguments.append("--" + k)
                    arguments.append(v)
                result = runner.invoke(add_member, arguments)
                debug_subprocess_pytest(result)
                assert append_to_roles_mock.call_count == 2 * (i + 1)
                append_to_roles_mock.assert_called_with(**test_case["expected_mutation_payload"])

    def test_update_member(self, mocker):
        TEST_CASES = [
            {
                "case_name": "AAU, when I update user's role with email adress, I see a success",
                "inputs": ["john.doe@test.com", "jane.doe@test.com"],
                "options": {"project-id": "project_id", "role": "REVIEWER"},
                "expected_mutation_payload": {
                    "role_id": "role_id_jane",
                    "project_id": "project_id",
                    "user_id": "user_id_jane",
                    "role": "REVIEWER",
                },
            },
            {
                "case_name": "AAU, when I update user's role from another project, I see a success",
                "inputs": [],
                "options": {
                    "project-id": "project_id",
                    "from-project": "project_id_source",
                },
                "expected_mutation_payload": {
                    "role_id": "role_id_jane",
                    "project_id": "project_id",
                    "user_id": "user_id_jane",
                    "role": "REVIEWER",
                },
            },
            {
                "case_name": "AAU, when I update user's role with csv file, I see a success",
                "inputs": [],
                "options": {
                    "project-id": "project_id",
                    "from-csv": "user_list.csv",
                    "role": "REVIEWER",
                },
                "expected_mutation_payload": {
                    "role_id": "role_id_jane",
                    "project_id": "project_id",
                    "user_id": "user_id_jane",
                    "role": "REVIEWER",
                },
            },
        ]
        runner = CliRunner()
        with runner.isolated_filesystem():
            # pylint: disable=unspecified-encodind
            with open("user_list.csv", "w") as f:
                writer = csv.writer(f)
                writer.writerow(["email"])
                writer.writerow(["john.doe@test.com"])
                writer.writerow(["jane.doe@test.com"])

            for i, test_case in enumerate(TEST_CASES):
                print(test_case["case_name"])
                arguments = test_case["inputs"]
                for k, v in test_case["options"].items():
                    arguments.append("--" + k)
                    arguments.append(v)
                result = runner.invoke(update_member, arguments)
                debug_subprocess_pytest(result)
                assert update_properties_in_role_mock.call_count == 2 * (i + 1)
                update_properties_in_role_mock.assert_called_with(
                    **test_case["expected_mutation_payload"]
                )

    def test_remove_member(self, mocker):
        TEST_CASES = [
            {
                "case_name": "AAU, when I remove users with email adress, I see a success",
                "inputs": ["john.doe@test.com", "jane.doe@test.com"],
                "options": {
                    "project-id": "project_id",
                },
                "expected_mutation_payload": {
                    "role_id": "role_id_jane",
                },
            },
            {
                "case_name": "AAU, when I remove all users, I see a success",
                "inputs": [],
                "options": {
                    "project-id": "project_id",
                },
                "flags": ["all"],
                "expected_mutation_payload": {
                    "role_id": "role_id_jane",
                },
            },
            {
                "case_name": "AAU, when I remove users with a  csv file, I see a success",
                "inputs": [],
                "options": {
                    "project-id": "project_id",
                    "from-csv": "user_list.csv",
                },
                "expected_mutation_payload": {
                    "role_id": "role_id_jane",
                },
            },
        ]
        runner = CliRunner()
        with runner.isolated_filesystem():
            # pylint: disable=unspecified-encodind
            with open("user_list.csv", "w") as f:
                writer = csv.writer(f)
                writer.writerow(["email"])
                writer.writerow(["john.doe@test.com"])
                writer.writerow(["jane.doe@test.com"])

            for i, test_case in enumerate(TEST_CASES):
                print(test_case["case_name"])
                arguments = test_case["inputs"]
                for k, v in test_case["options"].items():
                    arguments.append("--" + k)
                    arguments.append(v)
                if test_case.get("flags"):
                    arguments.extend(["--" + flag for flag in test_case["flags"]])
                result = runner.invoke(remove_member, arguments)
                debug_subprocess_pytest(result)
                assert delete_from_roles_mock.call_count == 2 * (i + 1)
                delete_from_roles_mock.assert_called_with(**test_case["expected_mutation_payload"])



"""Project module."""


class Project:  # pylint: disable=too-few-public-methods
    """
    Object that represents a project in Kili.
    It allows management operations such as uploading assets, uploading predictions,
    modifying the project's queue etc.
    It also allows queries from this project such as its assets, labels etc.
    """

    def __init__(self, client, project_id, input_type, title):
        self.project_id = project_id
        self.client = client
        self.title = title
        self.input_type = input_type



"""CLI's project member list subcommand"""

from typing import Dict, List, Optional, cast

import click
import pandas as pd
from tabulate import tabulate

from kili.cli.common_args import Arguments, Options
from kili.client import Kili

ROLE_ORDER = {v: i for i, v in enumerate(["ADMIN", "TEAM_MANAGER", "REVIEWER", "LABELER"])}


@click.command(name="list")
@Options.api_key
@Options.endpoint
@Arguments.project_id
@Options.tablefmt
def list_members(api_key: Optional[str], endpoint: Optional[str], project_id: str, tablefmt: str):
    """
    List the members of the project

    \b
    !!! Examples
        ```
        kili project member list <project_id> --stdout-format pretty
        ```

    """
    kili = Kili(api_key=api_key, api_endpoint=endpoint)
    users = cast(
        List[Dict],
        kili.project_users(
            project_id=project_id,
            fields=[
                "role",
                "activated",
                "user.email",
                "user.id",
                "user.firstname",
                "user.lastname",
                "user.organization.name",
            ],
            disable_tqdm=True,
        ),
    )
    users = pd.DataFrame(users)
    users = pd.concat([users.drop(["user"], axis=1), users["user"].apply(pd.Series)], axis=1)
    users = pd.concat(
        [users.drop(["organization"], axis=1), users["organization"].apply(pd.Series)],
        axis=1,
    )
    users = users.loc[users["activated"]]
    users.rename(
        columns={"role": "ROLE", "email": "EMAIL", "id": "ID", "name": "ORGANIZATION"},
        inplace=True,
    )
    users["NAME"] = users["firstname"].str.title() + " " + users["lastname"].str.title()
    users = users.sort_values(
        by=["ROLE", "lastname"],
        ascending=True,
        key=lambda column: column.map(ROLE_ORDER) if column.name == "ROLE" else column,
    )
    users = users[["ROLE", "NAME", "EMAIL", "ID", "ORGANIZATION"]]
    print(tabulate(users, headers="keys", tablefmt=tablefmt, showindex=False))



"""
Queries of label subscription
"""

from .fragments import LABEL_FRAGMENT

GQL_LABEL_CREATED_OR_UPDATED = f"""
subscription($projectID: ID!) {{
  data: labelCreatedOrUpdated(projectID: $projectID) {{
    {LABEL_FRAGMENT}
  }}
}}
"""



"""Exceptions of the package."""
import ast


class NotFound(Exception):
    """Used when a given object is not found in Kili"""

    def __init__(self, name: str):
        super().__init__()
        self.name = name

    def __str__(self):
        return f"Not found: '{self.name}'"


class AuthenticationFailed(Exception):
    """
    Used when the authentification fails
    """

    def __init__(self, api_key, api_endpoint):
        if api_key is None:
            super().__init__(
                "You need to provide an API KEY to connect."
                " Visit https://docs.kili-technology.com/reference/creating-an-api-key"
            )
        else:
            super().__init__(
                f"Connection to Kili endpoint {api_endpoint} failed with API key: {api_key}."
                " Check your connection and API key."
            )


class EndpointCompatibilityError(Exception):
    """
    Used when a resolver is called but is not compatible with the endpoint
    """

    def __init__(self, resolver, endpoint):
        super().__init__(
            f"Resolver {resolver} is not compatible with the following endpoint : {endpoint}"
        )


class GraphQLError(Exception):
    """
    Used when the GraphQL call returns an error
    """

    def __init__(self, error: Exception, batch_number=None):
        if batch_number is None:
            super().__init__(f'error: "{ast.literal_eval(str(error))[0]["message"]}"')
        else:
            super().__init__(
                f"error at index {100*batch_number}: "
                f'{ast.literal_eval(str(error))[0]["message"]}"'
            )



"""
Helpers for the project queries.
"""

from typing import List, Tuple


def get_project_url(project_id: str, api_endpoint: str):
    """
    Get the project url from the project id and the api_endpoint

    Args:
        project_id: the project id
        api_endpoint: Kili's API endpoint
    """
    domain = api_endpoint.replace("/api/label/v2/graphql", "")
    return domain + f"/label/projects/{project_id}/"


def get_project_metadata(project: dict, api_endpoint: str) -> List[Tuple]:
    """Get project metadata.

    Args:
        project: a project with its statistics fields

    Return:
        a list of metadata (key, value)
    """
    return [
        ("Title", project["title"]),
        ("Description", project["description"]),
        ("URL", get_project_url(project["id"], api_endpoint)),
    ]


def get_project_metrics(project: dict) -> Tuple[List[Tuple], List[Tuple]]:
    """Get project metrics.

    Args:
        project: a project with its statistics fields

    Return:
        dataset_metrics, quality_metrics: arrays that contain
        the project infos and progress metrics
    """

    if project["numberOfAssets"]:
        progress = round(
            (1 - project["numberOfRemainingAssets"] / project["numberOfAssets"]) * 100,
            1,
        )
    else:
        progress = 0

    dataset_metrics = [
        ("Total number of assets", project["numberOfAssets"]),
        ("Number of remaining assets", project["numberOfRemainingAssets"]),
        ("Skipped assets", project["numberOfAssetsWithSkippedLabels"]),
        ("Progress", str(progress) + "%"),
    ]
    quality_metrics = [
        ("Project consensus", project["consensusMark"] or "N/A"),
        ("Project honeypot", project["honeypotMark"] or "N/A"),
        ("Number of reviewed assets", project["numberOfReviewedAssets"]),
        ("Number of open issues", project["numberOfOpenIssues"]),
        ("Number of solved issues", project["numberOfSolvedIssues"]),
        ("Number of open questions", project["numberOfOpenQuestions"]),
        ("Number of solved questions", project["numberOfSolvedQuestions"]),
    ]
    return dataset_metrics, quality_metrics



"""Project user queries."""

from typing import Generator, List, Optional, Union

from typeguard import typechecked

from ...helpers import Compatible, format_result, fragment_builder
from ...types import ProjectUser
from ...utils.pagination import row_generator_from_paginated_calls
from .queries import GQL_PROJECT_USERS_COUNT, gql_project_users


class QueriesProjectUser:
    """Set of ProjectUser queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value,invalid-name
    @Compatible(["v1", "v2"])
    @typechecked
    def project_users(
        self,
        project_id: str,
        email: Optional[str] = None,
        id: Optional[str] = None,  # pylint: disable=redefined-builtin
        organization_id: Optional[str] = None,
        fields: List[str] = [
            "activated",
            "id",
            "role",
            "starred",
            "user.email",
            "user.id",
        ],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Return project users (possibly with their KPIs) that match a set of criteria


        Args:
            project_id: Identifier of the project
            email: Email of the user
            organization_id: Identifier of the user's organization
            fields: All the fields to request among the possible fields for the projectUsers
                See [the documentation](https://cloud.kili-technology.com/docs/python-graphql-api/graphql-api/#projectuser) for all possible fields.
            first: Maximum number of users to return
            skip: Number of project users to skip
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the project users is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            ```
            # Retrieve consensus marks of all users in project
            >>> kili.project_users(project_id=project_id, fields=['consensusMark', 'user.email'])
            ```
        """
        count_args = {
            "email": email,
            "id": id,
            "organization_id": organization_id,
            "project_id": project_id,
        }
        disable_tqdm = disable_tqdm or as_generator
        payload_query = {
            "where": {
                "id": id,
                "project": {
                    "id": project_id,
                },
                "user": {
                    "email": email,
                    "organization": {
                        "id": organization_id,
                    },
                },
            }
        }

        project_users_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_project_users,
            count_args,
            self._query_project_users,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return project_users_generator
        return list(project_users_generator)

    def _query_project_users(self, skip: int, first: int, payload: dict, fields: List[str]):

        payload.update({"skip": skip, "first": first})
        _gql_project_users = gql_project_users(fragment_builder(fields, ProjectUser))
        result = self.auth.client.execute(_gql_project_users, payload)
        return format_result("data", result)

    # pylint: disable=invalid-name
    @typechecked
    def count_project_users(
        self,
        project_id: str,
        email: Optional[str] = None,
        id: Optional[str] = None,  # pylint: disable=redefined-builtin
        organization_id: Optional[str] = None,
    ) -> int:
        """
        Counts the number of projects and their users that match a set of criteria

        Args:
            email: Email of the user
            organization_id: Identifier of the user's organization
            project_id: Identifier of the project

        Returns:
            The number of project users with the parameters provided
        """
        variables = {
            "where": {
                "id": id,
                "project": {
                    "id": project_id,
                },
                "user": {
                    "email": email,
                    "organization": {
                        "id": organization_id,
                    },
                },
            }
        }
        result = self.auth.client.execute(GQL_PROJECT_USERS_COUNT, variables)
        count = format_result("data", result)
        return count



"""Issue queries."""

from dataclasses import dataclass
from typing import Generator, List, Optional, Union

from typeguard import typechecked

from ...helpers import Compatible, format_result, fragment_builder
from ...types import Issue as IssueType
from ...utils.pagination import row_generator_from_paginated_calls
from .queries import GQL_ISSUES_COUNT, gql_issues


@dataclass
class QueriesIssue:
    """Set of Issue queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(["v1", "v2"])
    @typechecked
    def issues(
        self,
        project_id: str,
        fields: Optional[List[str]] = [
            "id",
            "createdAt",
            "hasBeenSeen",
            "issueNumber",
            "status",
            "type",
        ],
        first: Optional[int] = None,
        skip: Optional[int] = 0,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of issues that match a set of criteria.

        Args:
            project_id: Project ID the issue belongs to.
            fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#issue) for all possible fields.
            first: Maximum number of issues to return.
            skip: Number of issues to skip (they are ordered by their date of creation, first to last).
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the issues is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            >>> kili.issues(project_id=project_id, fields=['author.email']) # List all issues of a project and their authors
        """
        count_args = {"project_id": project_id}
        disable_tqdm = disable_tqdm or as_generator
        payload_query = {
            "where": {
                "project": {
                    "id": project_id,
                },
            },
        }

        issues_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_issues,
            count_args,
            self._query_issues,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return issues_generator
        return list(issues_generator)

    def _query_issues(self, skip: int, first: int, payload: dict, fields: List[str]):
        payload.update({"skip": skip, "first": first})
        _gql_issues = gql_issues(fragment_builder(fields, IssueType))
        result = self.auth.client.execute(_gql_issues, payload)
        return format_result("data", result)

    @Compatible(["v2"])
    @typechecked
    def count_issues(self, project_id: Optional[str] = None) -> int:
        """Count and return the number of api keys with the given constraints.

        Args:
            project_id: Project ID the issue belongs to.

        Returns:
            The number of issues with the parameters provided

        """
        variables = {
            "where": {
                "project": {
                    "id": project_id,
                },
            },
        }
        result = self.auth.client.execute(GQL_ISSUES_COUNT, variables)
        count = format_result("data", result)
        return count



"""Project version mutations."""

from dataclasses import dataclass
from typing import Optional

from typeguard import typechecked

from ...helpers import Compatible, format_result
from .queries import GQL_UPDATE_PROPERTIES_IN_PROJECT_VERSION


@dataclass
class MutationsProjectVersion:
    """Set of ProjectVersion mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @Compatible(["v2"])
    @typechecked
    def update_properties_in_project_version(self, project_version_id: str, content: Optional[str]):
        """Update properties of a project version.

        Args:
            project_version_id: Identifier of the project version
            content: Link to download the project version

        Returns:
            A result object which indicates if the mutation was successful.

        Examples:
            >>> kili.update_properties_in_project_version(
                    project_version_id=project_version_id,
                    content='test')
        """
        variables = {
            "content": content,
            "id": project_version_id,
        }
        result = self.auth.client.execute(GQL_UPDATE_PROPERTIES_IN_PROJECT_VERSION, variables)
        return format_result("data", result)



"""CLI's project member remove subcommand"""

import warnings
from typing import Optional

import click

from kili.cli.common_args import Arguments, Options, from_csv
from kili.cli.project.member.helpers import (
    check_exclusive_options,
    collect_members_from_csv,
    collect_members_from_emails,
    collect_members_from_project,
)
from kili.client import Kili


@click.command(name="rm")
@Options.api_key
@Options.endpoint
@Arguments.emails
@Options.project_id
@from_csv(["email"], ["role"])
@click.option(
    "--all",
    "all_members",
    type=bool,
    is_flag=True,
    default=False,
    help="Remove all users from project",
)
# pylint: disable=too-many-arguments
def remove_member(
    api_key: Optional[str],
    endpoint: Optional[str],
    emails: Optional[str],
    project_id: str,
    csv_path: Optional[str],
    all_members: bool,
):
    """Remove members from a Kili project

    Emails can be passed directly as arguments.
    You can provide several emails separated by spaces.

    \b
    !!! Examples
        ```
        kili project member rm \\
            --project-id <project_id> \\
            john.doe@test.com
        ```
        ```
        kili project member rm \\
            --project-id <project_id> \\
            --from-csv path/to/members.csv

        ```
        ```
        kili project member rm \\
            --project-id <project_id> \\
            --all
        ```
    """
    kili = Kili(api_key=api_key, api_endpoint=endpoint)

    check_exclusive_options(csv_path, None, emails, all_members)

    if csv_path is not None:
        members_to_rm = collect_members_from_csv(csv_path, None)
    elif all_members:
        members_to_rm = collect_members_from_project(kili, project_id, None)
    else:
        members_to_rm = collect_members_from_emails(emails, None)

    count = 0
    existing_members = kili.project_users(project_id=project_id, disable_tqdm=True)
    existing_members = {
        member["user"]["email"]: member["id"] for member in existing_members if member["activated"]
    }

    for member in members_to_rm:
        email = member["email"]
        if email in existing_members.keys():
            kili.delete_from_roles(role_id=existing_members[email])
            count += 1
        else:
            warnings.warn(f"{email} is not an active member of the project.")

    print(f"{count} users have been successfully removed from project: {project_id}")



"""
Tests for the creation and query of a project from the client
"""


import json

import pytest

from kili.client import Kili
from kili.exceptions import NotFound
from kili.project import Project


def mocked__projects(project_id, **_):
    """Simulates a project call and return an object detection project if
    the given project_id is the one of the fixture."""
    project_fixture = json.load(open("test/fixtures/object_detection_project_fixture.json"))
    if project_id == project_fixture[0]["id"]:
        return json.load(open("test/fixtures/object_detection_project_fixture.json"))
    else:
        return []


class TestGetProject:
    """
    test the get_project function
    """

    def test_get_unexisting_project(self, mocker):
        """Test that get_project on a unexisting project_id raises a NotFound Error."""
        mocker.patch("kili.client.Kili.__init__", return_value=None)
        mocker.patch("kili.client.Kili.projects", side_effect=mocked__projects)
        kili = Kili()
        with pytest.raises(NotFound):
            kili.get_project("this_project_id_do_not_exist")

    def test_get_existing_project(self, mocker):
        """Test that get_project on an existing project_id return a Project with the right id."""
        mocker.patch("kili.client.Kili.__init__", return_value=None)
        mocker.patch("kili.client.Kili.projects", side_effect=mocked__projects)
        kili = Kili()
        project = kili.get_project("cl0wihlop3rwc0mtj9np28ti2")
        assert isinstance(project, Project)
        assert project.project_id == "cl0wihlop3rwc0mtj9np28ti2"



"""
Queries of notification queries
"""


def gql_notifications(fragment):
    """
    Return the GraphQL notifications query
    """
    return f"""
query ($where: NotificationWhere!, $first: PageSize!, $skip: Int!) {{
  data: notifications(where: $where, first: $first, skip: $skip) {{
    {fragment}
  }}
}}
"""


GQL_NOTIFICATIONS_COUNT = """
query($where: NotificationWhere!) {
  data: countNotifications(where: $where)
}
"""



"""CLI's project import subcommand"""

import os
import urllib.request
from typing import Dict, List, Optional, Tuple, cast

import click
from typeguard import typechecked

from kili.cli.common_args import Arguments, Options, from_csv
from kili.cli.helpers import (
    check_exclusive_options,
    collect_from_csv,
    get_external_id_from_file_path,
)
from kili.client import Kili
from kili.exceptions import NotFound
from kili.helpers import file_check_function_from_input_type, get_file_paths_to_upload
from kili.mutations.asset.helpers import generate_json_metadata_array

# pylint: disable=consider-using-with


def type_check_asset(key, value):
    """type check value based on key"""
    if (
        key == "content"
        and not os.path.isfile(value)
        and not urllib.request.urlopen(value).getcode() == 200
    ):
        return f"{value} is not a valid url or path to a file."

    return ""


@click.command(name="import")
@Arguments.files
@Options.api_key
@Options.endpoint
@Options.project_id
@from_csv(["external_id", "content"], [])
@click.option(
    "--frames",
    "as_frames",
    type=bool,
    default=False,
    is_flag=True,
    help="Only for a frame project, import videos as frames. "
    "The import time is longer with this option.",
)
@click.option(
    "--fps", type=int, help="Only for a frame project, import videos with a specific frame rate"
)
@Options.verbose
@typechecked
# pylint: disable=too-many-arguments
def import_assets(
    api_key: Optional[str],
    endpoint: Optional[str],
    project_id: str,
    files: Optional[Tuple[str, ...]],
    csv_path: Optional[str],
    fps: Optional[int],
    as_frames: bool,
    verbose: bool,
):
    """
    Add assets into a project

    Files can be paths to files or to folders. You can provide several paths separated by spaces.

    If no Files are provided, --from-csv can be used to import
    assets from a CSV file with two columns:

    - `external_id`: external id of the asset.
    - `content`: paths to the asset file or a url hosting the asset.

    \b
    !!! Examples
        ```
        kili project import \\
            dir1/dir2/ dir1/dir3/test1.png \\
            --project-id <project_id>
        ```
        ```
        kili project import \\
            dir1/dir3/video.mp4 \\
            --project-id <project_id> \\
            --frames \\
            --fps 24
        ```
        ```
        kili project import \\
            --from-csv assets_list.csv \\
            --project-id <project_id> \\
            --frames \\
            --fps 24
        ```
    \b
    !!! warning "Unsupported imports"
        Currently, this command does not support:

        - the import of videos from local frames, rich text and time series assets
        - the import of assets with metadata or with a custom external_id

        For such imports, please use the `append_many_to_dataset` method in the Kili SDK.
    """
    kili = Kili(api_key=api_key, api_endpoint=endpoint)
    try:
        input_type = cast(
            List[Dict], kili.projects(project_id, disable_tqdm=True, fields=["inputType"])
        )[0]["inputType"]
    except:
        # pylint: disable=raise-missing-from
        raise NotFound(f"project ID: {project_id}")

    if input_type not in ("FRAME", "VIDEO") and (fps is not None or as_frames is True):
        illegal_option = "fps and frames are"
        if not as_frames:
            illegal_option = "fps is"
        if fps is None:
            illegal_option = "frames is"
        raise ValueError(f"{illegal_option} only valid for a VIDEO project")

    check_exclusive_options(csv_path, files)

    if len(files) > 0:
        files_to_upload = get_file_paths_to_upload(
            files, file_check_function_from_input_type(input_type), verbose
        )
        if len(files_to_upload) == 0:
            raise ValueError(
                "No files to upload. "
                "Check that the paths exist and file types are compatible with the project"
            )
        external_ids = [get_external_id_from_file_path(path) for path in files_to_upload]

    elif csv_path is not None:
        row_dict = collect_from_csv(
            csv_path=csv_path,
            required_columns=["external_id", "content"],
            optional_columns=[],
            type_check_function=type_check_asset,
        )

        files_to_upload = [row["content"] for row in row_dict]
        external_ids = [row["external_id"] for row in row_dict]

        if len(files_to_upload) == 0:
            raise ValueError(f"No valid asset files or url were found in csv: {csv_path}")

    json_metadata_array = generate_json_metadata_array(
        as_frames, fps, len(files_to_upload), input_type
    )

    kili.append_many_to_dataset(
        project_id=project_id,
        content_array=files_to_upload,
        external_id_array=external_ids,
        json_metadata_array=json_metadata_array,
    )

    if as_frames:
        print(
            f"The import of {len(files_to_upload)} files have just started, "
            "you will receive a notification as soon as it is ready."
        )
    else:
        print(f"{len(files_to_upload)} files have been successfully imported")



"""
Api key queries
"""

from typing import Generator, List, Optional, Union

from typeguard import typechecked

from ...helpers import Compatible, format_result, fragment_builder
from ...types import ApiKey as ApiKeyType
from ...utils.pagination import row_generator_from_paginated_calls
from .queries import GQL_API_KEYS_COUNT, gql_api_keys


class QueriesApiKey:
    """
    Set of ApiKey queries
    """

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(["v2"])
    @typechecked
    def api_keys(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
        skip: int = 0,
        fields: List[str] = ["id", "name", "createdAt", "revoked"],
        first: Optional[int] = None,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of API keys that match a set of constraints.

        !!! info
            You can only query your own API keys

        Args:
            api_key_id: Identifier of the API key to retrieve.
            user_id: Identifier of the user.
            api_key: Value of the API key.
            skip: Number of assets to skip (they are ordered by their date of creation, first to last).
            fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://cloud.kili-technology.com/docs/python-graphql-api/graphql-api/#apikey) for all possible fields.
            first: Maximum number of API keys to return.
            disable_tqdm: If `True`, the progress bar will be disabled.
            as_generator: If `True`, a generator on the API key is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.


        Examples:
            >>> kili.api_keys(user_id=user_id)
            >>> kili.api_keys(api_key=api_key)
            >>> kili.api_keys(api_key=api_key, as_generator=False)
        """

        saved_args = locals()
        count_args = {
            k: v for (k, v) in saved_args.items() if k in ["user_id", "api_key_id", "api_key"]
        }
        disable_tqdm = disable_tqdm or as_generator

        payload_query = {
            "where": {
                "user": {"id": user_id, "apiKey": api_key},
                "id": api_key_id,
            },
        }

        api_keys_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_api_keys,
            count_args,
            self._query_api_keys,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return api_keys_generator
        return list(api_keys_generator)

    def _query_api_keys(self, skip: int, first: int, payload: dict, fields: List[str]):

        payload.update({"skip": skip, "first": first})
        _gql_api_keys = gql_api_keys(fragment_builder(fields, ApiKeyType))
        result = self.auth.client.execute(_gql_api_keys, payload)
        return format_result("data", result)

    @Compatible(["v2"])
    @typechecked
    def count_api_keys(
        self,
        api_key_id: Optional[str] = None,
        user_id: Optional[str] = None,
        api_key: Optional[str] = None,
    ) -> int:
        """Count and return the number of api keys with the given constraints.

        Args:
            api_key_id: Identifier of the API key to retrieve.
            user_id: Identifier of the user.
            api_key: Value of the api key.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            >>> kili.count_api_keys(user_id=user_id)
            3
            >>> kili.count_api_keys(api_key=api_key)
            1
        """
        variables = {
            "where": {
                "user": {"id": user_id, "apiKey": api_key},
                "id": api_key_id,
            },
        }
        result = self.auth.client.execute(GQL_API_KEYS_COUNT, variables)
        count = format_result("data", result)
        return count



"""Tests for utils module"""

from kili.utils.pagination import (
    batch_iterator_builder,
    batch_object_builder,
    row_generator_from_paginated_calls,
)

from .utils import mocked_count_method, mocked_query_method

TEST_CASES = [
    {
        "case": "AAU, When I query objects with first=50, I get the first 50 objects",
        "args": {"first": 50},
        "expected_result": ({"id": i} for i in range(50)),
    },
    {
        "case": "AAU, When I query objects with first=200, I get the first 200 objects",
        "args": {"first": 200},
        "expected_result": ({"id": i} for i in range(200)),
    },
    {
        "case": "AAU, When I query objects with first=50 and skip=20, I skip the 20"
        " first objects I get the 50 objects following",
        "args": {"first": 50, "skip": 20},
        "expected_result": ({"id": i} for i in range(20, 70)),
    },
    {
        "case": "AAU, When I query objects with first=50 and skip=20, I skip the 20"
        "first objects I get the 200 following objects",
        "args": {"first": 200, "skip": 20},
        "expected_result": ({"id": i} for i in range(20, 220)),
    },
    {
        "case": "AAU, When I query objects with first=25100, I get the first 25100 objects",
        "args": {"first": 25100},
        "expected_result": ({"id": i} for i in range(25100)),
    },
    {
        "case": "AAU, When I query objects with first=20 and disable_tqdm, I get the first 2O objects",
        "args": {"first": 20, "disable_tqdm": True},
        "expected_result": ({"id": i} for i in range(20)),
    },
    {
        "case": "AAU, When I query objects with first=20, skip=20 and disable_tqdm, I  skip the 20"
        "first objects and I get the 20 following objects",
        "args": {"first": 20, "skip": 20, "disable_tqdm": True},
        "expected_result": ({"id": i} for i in range(20, 40)),
    },
    {
        "case": "AAU, When I query objects with first=0, I get no objects",
        "args": {"first": 0},
        "expected_result": iter(()),
    },
]


def test_row_generator_from_paginated_calls():
    """
    Simulates a count query result by returning a list of ids
    """
    for test_case in TEST_CASES:
        case_name = test_case["case"]
        expected = test_case["expected_result"]
        skip = test_case["args"].get("skip", 0)
        first = test_case["args"].get("first")
        disable_tqdm = test_case["args"].get("disable_tqdm", False)

        actual = row_generator_from_paginated_calls(
            skip,
            first,
            mocked_count_method,
            {},
            mocked_query_method,
            {first, skip},
            [],
            disable_tqdm,
        )
        assert all(a == b for a, b in zip(actual, expected)), f'Test case "{case_name}" failed'
        assert type(actual).__name__ == "generator", f'Test case "{case_name}" failed'


def test_batch_iterator_builder():
    """Test batch iterator builder."""
    TEST_CASE = [
        {
            "case": "When the size of the iterable is not a multiple of the batch_size, I see a success",
            "batch_size": 3,
            "iterable": [i for i in range(7)],
            "expected_result": (i for i in [[0, 1, 2], [3, 4, 5], [6]]),
        }
    ]
    for test_case in TEST_CASE:
        actual = batch_iterator_builder(test_case["iterable"], batch_size=test_case["batch_size"])
        expected = test_case["expected_result"]
        case_name = test_case["case"]
        assert all(a == b for a, b in zip(actual, expected)), f'Test case "{case_name}" failed'


def test_batch_object_builder():
    """Test batch iterator builder for several arrays in the same time."""
    array1 = [i for i in range(10)]
    array2 = [i for i in range(10, 20)]
    array3 = [i for i in range(20, 30)]
    TEST_CASE = [
        {
            "case": "When I have one property I see that it returns paginated batches",
            "batch_size": 3,
            "properties_to_batch": {"a1": array1},
            "expected_result": (
                i
                for i in [
                    {"a1": [0, 1, 2]},
                    {"a1": [3, 4, 5]},
                    {"a1": [6, 7, 8]},
                    {"a1": [9]},
                ]
            ),
        },
        {
            "case": "When I have several iterables I see it creates a common batch",
            "batch_size": 4,
            "properties_to_batch": {"a1": array1, "a2": array2, "a3": array3},
            "expected_result": (
                i
                for i in [
                    {
                        "a1": [0, 1, 2, 3],
                        "a2": [10, 11, 12, 13],
                        "a3": [20, 21, 22, 23],
                    },
                    {
                        "a1": [4, 5, 6, 7],
                        "a2": [14, 15, 16, 17],
                        "a3": [24, 25, 26, 27],
                    },
                    {"a1": [8, 9], "a2": [18, 19], "a3": [28, 29]},
                ]
            ),
        },
        {
            "case": "When I have None arrays, I see that it stays None in the batches",
            "batch_size": 4,
            "properties_to_batch": {"a1": array1, "a2": None, "a3": array2, "a4": None},
            "expected_result": (
                i
                for i in [
                    {
                        "a1": [0, 1, 2, 3],
                        "a2": None,
                        "a3": [10, 11, 12, 13],
                        "a4": None,
                    },
                    {
                        "a1": [4, 5, 6, 7],
                        "a2": None,
                        "a3": [14, 15, 16, 17],
                        "a4": None,
                    },
                    {"a1": [8, 9], "a2": None, "a3": [18, 19], "a4": None},
                ]
            ),
        },
        {
            "case": "When I have only None, I see it return one batch of None",
            "batch_size": 2,
            "properties_to_batch": {"a1": None, "a2": None, "a3": None},
            "expected_result": (i for i in [{"a1": None, "a2": None, "a3": None}]),
        },
    ]
    for test_case in TEST_CASE:
        actual = batch_object_builder(test_case["properties_to_batch"], test_case["batch_size"])
        expected = test_case["expected_result"]
        case_name = test_case["case"]
        assert all(a == b for a, b in zip(actual, expected)), f'Test case "{case_name}" failed'



"""
Fragments of project mutations
"""

PROJECT_FRAGMENT = """
id
jsonInterface
title
roles {
    user {
      id
      email
    }
    role
}
"""

PROJECT_FRAGMENT_ID = """
id
"""

PROJECT_FRAGMENT_PUBLIC_TOKEN = """
id
publicToken
"""

ROLE_FRAGMENT = """
user {
  id
  email
}
role
"""



"""Organization queries."""

from datetime import datetime
from typing import Generator, List, Optional, Union

from typeguard import typechecked

from ...helpers import Compatible, format_result, fragment_builder
from ...types import Organization
from ...utils.pagination import row_generator_from_paginated_calls
from .queries import (
    GQL_ORGANIZATION_METRICS,
    GQL_ORGANIZATIONS_COUNT,
    gql_organizations,
)


class QueriesOrganization:
    """
    Set of Organization queries
    """

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(["v1", "v2"])
    @typechecked
    def organizations(
        self,
        email: Optional[str] = None,
        organization_id: Optional[str] = None,
        fields: List[str] = ["id", "name"],
        first: Optional[int] = None,
        skip: int = 0,
        disable_tqdm: bool = False,
        as_generator: bool = False,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a generator or a list of organizations that match a set of criteria.

        Args:
            email : Email of a user of the organization
            organization_id : Identifier of the organization
            fields: All the fields to request among the possible fields for the organizations.
                See [the documentation](https://cloud.kili-technology.com/docs/python-graphql-api/graphql-api/#organization) for all possible fields.
            first: Maximum number of organizations to return.
            Skip: Number of skipped organizations (they are ordered by creation date)
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the organizations is returned.

        Returns:
            A result object which contains the query if it was successful,
                or an error message.

        Examples:
            >>> kili.organizations(organization_id=organization_id, fields=['users.email'])
            [{'users': [{'email': 'john@doe.com'}]}]
        """

        count_args = {"email": email, "organization_id": organization_id}
        disable_tqdm = disable_tqdm or as_generator

        payload_query = {
            "where": {
                "id": organization_id,
                "user": {
                    "email": email,
                },
            }
        }

        organizations_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_organizations,
            count_args,
            self._query_organizations,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return organizations_generator
        return list(organizations_generator)

    def _query_organizations(self, skip: int, first: int, payload: dict, fields: List[str]):

        payload.update({"skip": skip, "first": first})
        _gql_organizations = gql_organizations(fragment_builder(fields, Organization))
        result = self.auth.client.execute(_gql_organizations, payload)
        return format_result("data", result)

    @Compatible(["v2"])
    @typechecked
    def count_organizations(
        self, email: Optional[str] = None, organization_id: Optional[str] = None
    ) -> int:
        """Count organizations that match a set of criteria.

        Args:
            email: Email of a user of the organization
            organization_id: Identifier of the organization

        Returns:
            A result object which contains the query if it was successful,
                or an error message.
        """
        variables = {
            "where": {
                "id": organization_id,
                "user": {
                    "email": email,
                },
            }
        }
        result = self.auth.client.execute(GQL_ORGANIZATIONS_COUNT, variables)
        return format_result("data", result)

    @Compatible(["v2"])
    @typechecked
    def organization_metrics(
        self,
        organization_id: str,
        start_date: datetime,
        end_date: datetime,
    ):
        """Get organization metrics.

        Args:
            organization_id: Identifier of the organization
            start_date: Start date of the metrics computation
            end_date: End date of the metrics computation

        Returns:
            A result object which contains the query if it was successful,
                or an error message.
        """
        if start_date is None:
            start_date = datetime.now()
        if end_date is None:
            end_date = datetime.now()
        variables = {
            "where": {
                "organizationId": organization_id,
                "startDate": start_date.isoformat(sep="T", timespec="milliseconds") + "Z",
                "endDate": end_date.isoformat(sep="T", timespec="milliseconds") + "Z",
            }
        }
        result = self.auth.client.execute(GQL_ORGANIZATION_METRICS, variables)
        return format_result("data", result)



"""
Test mutations with pytest
"""
import os
import shutil
import tempfile
import unittest
import uuid

import pytest
import requests

from kili.mutations.asset.helpers import (
    get_file_mimetype,
    process_append_many_to_dataset_parameters,
    process_content,
)
from kili.mutations.asset.queries import GQL_APPEND_MANY_FRAMES_TO_DATASET


class LocalDownloader:
    def __init__(self, directory):
        self.directory = directory

    def __call__(self, url):
        content = requests.get(url)
        name = os.path.basename(url)
        path = os.path.join(self.directory, f"{str(uuid.uuid4())}-{name}")
        with open(path, "wb") as file:
            file.write(content.content)
        return path


class TestMimeType:
    """
    Tests if the mime type is the correct one
    """

    def should_have_right_mimetype(self, content_array, json_content_array, expected_mimetype):
        mimetype = get_file_mimetype(content_array, json_content_array)
        assert mimetype == expected_mimetype, f"Bad mimetype {mimetype}"

    def test_contents_empty(self):
        content_array = None
        json_content_array = None
        self.should_have_right_mimetype(content_array, json_content_array, None)

    def test_mimetype_url(self, tmpdir):
        url = "https://storage.googleapis.com/label-public-staging/car/car_1.jpg"
        content_array = [url]
        json_content_array = None
        self.should_have_right_mimetype(content_array, json_content_array, None)

    def test_mimetype_image(self, tmpdir):
        url = "https://storage.googleapis.com/label-public-staging/car/car_1.jpg"
        downloader = LocalDownloader(tmpdir)
        path = downloader(url)
        content_array = [path]
        json_content_array = None
        self.should_have_right_mimetype(content_array, json_content_array, "image/jpeg")

    def test_mimetype_geotiff(self, tmpdir):
        url = "https://storage.googleapis.com/label-public-staging/geotiffs/bogota.tif"
        downloader = LocalDownloader(tmpdir)
        path = downloader(url)
        content_array = [path]
        json_content_array = None
        self.should_have_right_mimetype(content_array, json_content_array, "image/tiff")

    def test_cannot_upload_mp4_to_image_project(self):
        path = "./test.mp4"
        content_array = [path]
        json_content_array = None
        processed_content = process_content("IMAGE", content_array, json_content_array)
        assert processed_content == [None]

    def test_cannot_upload_png_to_frame_project(self):
        path = "./test.png"
        content_array = [path]
        json_content_array = None
        processed_content = process_content("VIDEO", content_array, json_content_array)
        assert processed_content == [None]

    def test_cannot_upload_text_to_pdf_project(self):
        path = "Hello world"
        content_array = [path]
        json_content_array = None
        processed_content = process_content("PDF", content_array, json_content_array)
        assert processed_content == [None]

    def test_can_upload_png_to_image_project(self):
        with pytest.raises(FileNotFoundError):
            path = "./test.png"
            content_array = [path]
            json_content_array = None
            process_content("IMAGE", content_array, json_content_array)


class TestUploadTiff(unittest.TestCase):
    """
    Test functions for uploading geotiff
    """

    def setUp(self):
        # Create a temporary directory
        self.test_dir = tempfile.mkdtemp()

    def tearDown(self):
        # Remove the directory after the test
        shutil.rmtree(self.test_dir)

    def test_geotiff_upload_properties(self):
        url = "https://storage.googleapis.com/label-public-staging/geotiffs/bogota.tif"
        downloader = LocalDownloader(self.test_dir)
        path = downloader(url)
        input_type = "IMAGE"
        content_array = [path]
        external_id_array = ["bogota"]
        is_honeypot_array = None
        status_array = None
        json_content_array = None
        json_metadata_array = None
        properties, upload_type, request = process_append_many_to_dataset_parameters(
            input_type,
            content_array,
            external_id_array,
            is_honeypot_array,
            status_array,
            json_content_array,
            json_metadata_array,
        )
        assert properties["json_metadata_array"] == ["{}"]
        assert properties["content_array"][0].startswith("data:image/tiff;base64,SUkqAAgABABre")
        assert properties["external_id_array"] == ["bogota"]
        assert upload_type == "GEO_SATELLITE", "uploadType do not match"
        assert request == GQL_APPEND_MANY_FRAMES_TO_DATASET, "Requests do not match"



"""CLI's project label subcommand"""

import json
import os
import warnings
from typing import Optional, Tuple

import click

from kili.cli.common_args import Arguments, Options, from_csv
from kili.cli.helpers import (
    check_exclusive_options,
    collect_from_csv,
    get_external_id_from_file_path,
)
from kili.client import Kili
from kili.exceptions import NotFound
from kili.helpers import get_file_paths_to_upload
from kili.mutations.label.helpers import generate_create_predictions_arguments


def type_check_label(key, value):
    """type check value based on key"""
    if key == "json_response_path" and not (os.path.isfile(value) and value.endswith(".json")):
        return f"{value} is not a valid path to a json file, "

    return ""


@click.command(name="label")
@Options.api_key
@Options.endpoint
@Arguments.files
@from_csv(["external_id", "json_response_path"], [])
@Options.project_id
@click.option(
    "--prediction",
    "is_prediction",
    type=bool,
    is_flag=True,
    default=False,
    help="Tells to import labels as predictions, which means that they will appear "
    "as pre-annotations in the Kili interface",
)
@click.option(
    "--model-name",
    type=str,
    help="Name of the model that generated predictions, " "if labels are sent as predictions",
)
@Options.verbose
# pylint: disable=too-many-arguments, too-many-locals
def import_labels(
    api_key: Optional[str],
    endpoint: Optional[str],
    files: Optional[Tuple[str, ...]],
    csv_path: str,
    project_id: str,
    is_prediction: bool,
    model_name: str,
    verbose: bool,
):
    """
    Import labels or predictions

    Files can be paths to files or to folders. You can provide several paths separated by spaces.
    The labels to import have to be in the Kili format and stored in a json file.
    File's name must be equal to asset's external_id.

    If no files are provided, --from-csv can be used to import
    assets from a CSV file with two columns:
    - `external_id`: external id for which you want to import labels.
    - `json_response_path`: paths to the json files containing the json_response to upload.

    \b
    !!! Examples "CSV file template"
        ```
        external_id,json_response_path
        asset1,./labels/label_asset1.json
        asset2,./labels/label_asset2.json
        ```

    \b
    !!! Examples
        To import default labels:
        ```
        kili project label \\
             dir1/dir2/ dir1/dir3/test1.json \\
            --project-id <project_id>
        ```
        ```
        kili project label \\
            --from-csv path/to/file.csv \\
            --project-id <project_id>
        ```
        To import labels as predictions:
        ```
        kili project label \\
            --from-csv path/to/file.csv \\
            --project-id <project_id> \\
            --prediction \\
            --model-name YOLO-run-3
        ```

    """
    if is_prediction and model_name is None:
        raise ValueError(
            "When importing labels as prediction, "
            "you must provide a model name with the --model-name option"
        )

    check_exclusive_options(csv_path, files)

    kili = Kili(api_key=api_key, api_endpoint=endpoint)

    if kili.count_projects(project_id=project_id) == 0:
        raise NotFound(f"project ID: {project_id}")

    if len(files) > 0:
        label_paths = get_file_paths_to_upload(files, lambda path: path.endswith(".json"), verbose)
        if len(label_paths) == 0:
            raise ValueError(
                "No label files to upload. " "Check that the paths exist and file types are .json"
            )
        external_ids = [get_external_id_from_file_path(path) for path in label_paths]

    elif csv_path is not None:

        labels_to_add = collect_from_csv(
            csv_path=csv_path,
            required_columns=["external_id", "json_response_path"],
            optional_columns=[],
            type_check_function=type_check_label,
        )

        if len(labels_to_add) == 0:
            raise ValueError(f"No json files were found in csv: {csv_path}")

        label_paths = [label["json_response_path"] for label in labels_to_add]
        external_ids = [label["external_id"] for label in labels_to_add]

    asset_in_project_external_ids = kili.assets(
        project_id=project_id, fields=["externalId"], disable_tqdm=True
    )
    asset_in_project_external_ids = set(
        asset["externalId"] for asset in asset_in_project_external_ids
    )

    label_index_to_import = []
    for i, external_id in enumerate(external_ids):
        if external_id in asset_in_project_external_ids:
            label_index_to_import.append(i)
        else:
            warnings.warn(f"{external_id} is not an asset of project ID: {project_id}.")

    if is_prediction:
        create_predictions_arguments = generate_create_predictions_arguments(
            [label_paths[i] for i in label_index_to_import],
            [external_ids[i] for i in label_index_to_import],
            model_name,
            project_id,
        )
        kili.create_predictions(**create_predictions_arguments)
        print(f"{len(external_ids)} labels have been successfully imported")

    else:
        for i in label_index_to_import:
            with open(label_paths[i], encoding="utf-8") as label_file:
                json_response = json.load(label_file)

            kili.append_to_labels(
                label_asset_external_id=external_ids[i],
                json_response=json_response,
                project_id=project_id,
            )

    print(f"{len(label_index_to_import)} labels have been successfully imported")



# pylint: disable=invalid-name,too-many-instance-attributes

"""
GraphQL types
"""

from dataclasses import dataclass


@dataclass
class License:
    """
    A Wrapper for License GraphQL object.
    """

    api = "api"
    apiPriority = "apiPriority"
    enableSmartTools = "enableSmartTools"
    expiryDate = "expiryDate"
    inputType = "inputType"
    maxNumberOfLabeledAssets = "maxNumberOfLabeledAssets"
    publicProject = "publicProject"
    seats = "seats"
    startDate = "startDate"
    type = "type"
    uploadCloudData = "uploadCloudData"
    organizationId = "organizationId"


@dataclass
class OrganizationWithoutUser:
    """
    A wrapper for Organization GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id = "id"
    address = "address"
    canSeeDataset = "canSeeDataset"
    city = "city"
    country = "country"
    license = License
    name = "name"
    numberOfAnnotations = "numberOfAnnotations"
    numberOfLabeledAssets = "numberOfLabeledAssets"
    numberOfHours = "numberOfHours"
    zipCode = "zipCode"


@dataclass
class UserWithoutProjectUsers:
    """
    A wrapper for User GraphQL object.
    """

    id = "id"
    activated = "activated"
    createdAt = "createdAt"
    email = "email"
    firstname = "firstname"
    lastname = "lastname"
    organization = OrganizationWithoutUser
    organizationId = "organizationId"
    organizationIdCompute = "organizationIdCompute"
    organizationRole = "organizationRole"
    rights = "rights"
    updatedAt = "updatedAt"


@dataclass
class Organization(OrganizationWithoutUser):
    """
    A wrapper for Organization GraphQL object.
    """

    users = UserWithoutProjectUsers


@dataclass
class ProjectUserWithoutProject:
    """
    A wrapper for ProjectUser GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id = "id"
    activated = "activated"
    consensusMark = "consensusMark"
    consensusMarkPerCategory = "consensusMarkPerCategory"
    consensusMarkCompute = "consensusMarkCompute"
    honeypotMark = "honeypotMark"
    honeypotMarkCompute = "honeypotMarkCompute"
    lastLabelingAt = "lastLabelingAt"
    lastLabelingAtCompute = "lastLabelingAtCompute"
    numberOfAnnotations = "numberOfAnnotations"
    numberOfAnnotationsCompute = "numberOfAnnotationsCompute"
    numberOfLabeledAssets = "numberOfLabeledAssets"
    numberOfLabels = "numberOfLabels"
    numberOfLabelsCompute = "numberOfLabelsCompute"
    numberOfLabeledAssets = "numberOfLabeledAssets"
    numberOfLabeledAssetsCompute = "numberOfLabeledAssetsCompute"
    role = "role"
    starred = "starred"
    totalDuration = "totalDuration"
    totalDurationCompute = "totalDurationCompute"
    user = UserWithoutProjectUsers


@dataclass
class ProjectWithoutDataset:
    """
    A wrapper for Project GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id = "id"
    assetMetadata = "assetMetadata"
    assetMetadataCompute = "assetMetadataCompute"
    author = UserWithoutProjectUsers
    consensusMark = "consensusMark"
    consensusMarkPerCategory = "consensusMarkPerCategory"
    consensusTotCoverage = "consensusTotCoverage"
    createdAt = "createdAt"
    description = "description"
    honeypotMark = "honeypotMark"
    inputType = "inputType"
    instructions = "instructions"
    interface = "interface"
    interfaceCompute = "interfaceCompute"
    jsonInterface = "jsonInterface"
    minConsensusSize = "minConsensusSize"
    mlTasks = "mlTasks"
    mlTasksCompute = "mlTasksCompute"
    numberOfRemainingAssets = "numberOfRemainingAssets"
    numberOfAssets = "numberOfAssets"
    numberOfAssetsWithSkippedLabels = "numberOfAssetsWithSkippedLabels"
    numberOfOpenIssues = "numberOfOpenIssues"
    numberOfOpenQuestions = "numberOfOpenQuestions"
    numberOfSolvedIssues = "numberOfSolvedIssues"
    numberOfSolvedQuestions = "numberOfSolvedQuestions"
    numberOfReviewedAssets = "numberOfReviewedAssets"
    readPermissionsForAssetsAndLabels = "readPermissionsForAssetsAndLabels"
    reviewCoverage = "reviewCoverage"
    rights = "rights"
    roles = ProjectUserWithoutProject
    shouldRelaunchKpiComputation = "shouldRelaunchKpiComputation"
    title = "title"
    updatedAt = "updatedAt"
    useHoneyPot = "useHoneyPot"


@dataclass
class ProjectUser(ProjectUserWithoutProject):
    """
    A wrapper for ProjectUser GraphQL object.
    """

    project = ProjectWithoutDataset


@dataclass
class UserWithoutApiKey(UserWithoutProjectUsers):
    """
    A wrapper for User GraphQL object.
    """

    projectUsers = ProjectUser


@dataclass
class ApiKey:
    """
    A wrapper for ApiKey GraphQL object.
    """

    createdAt = "createdAt"
    id = "id"
    key = "key"
    name = "name"
    revoked = "revoked"
    user = UserWithoutApiKey
    userId = "userId"


@dataclass
class User(UserWithoutApiKey):
    """
    A wrapper for User GraphQL object.
    """

    apiKeys = ApiKey


@dataclass
class LabelWithoutLabelOf:
    """
    A wrapper for Label GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id = "id"
    assetIdCompute = "assetIdCompute"
    authorIdCompute = "authorIdCompute"
    author = User
    createdAt = "createdAt"
    honeypotMark = "honeypotMark"
    honeypotMarkCompute = "honeypotMarkCompute"
    inferenceMark = "inferenceMark"
    inferenceMarkCompute = "inferenceMarkCompute"
    isLatestLabelForUser = "isLatestLabelForUser"
    isLatestLabelForUserCompute = "isLatestLabelForUserCompute"
    isLatestDefaultLabelForUser = "isLatestDefaultLabelForUser"
    isLatestDefaultLabelForUserCompute = "isLatestDefaultLabelForUserCompute"
    isLatestReviewLabelForUser = "isLatestReviewLabelForUser"
    isLatestReviewLabelForUserCompute = "isLatestReviewLabelForUserCompute"
    jsonResponse = "jsonResponse"
    labelType = "labelType"
    modelName = "modelName"
    numberOfAnnotations = "numberOfAnnotations"
    numberOfAnnotationsCompute = "numberOfAnnotationsCompute"
    projectIdCompute = "projectIdCompute"
    responseCompute = "responseCompute"
    searchCompute = "searchCompute"
    secondsToLabel = "secondsToLabel"
    skipped = "skipped"
    totalSecondsToLabel = "totalSecondsToLabel"
    totalSecondsToLabelCompute = "totalSecondsToLabelCompute"


@dataclass
class Lock:
    """
    A wrapper for Lock GraphQL object.
    """

    id = "id"
    author = User
    authorIdCompute = "authorIdCompute"
    createdAt = "createdAt"
    lockType = "lockType"
    lockOfIdCompute = "lockOfIdCompute"


@dataclass
class CommentsWithoutCommentsOf:
    """
    A wrapper for Comment GraphQL object.
    Defined in two steps to avoid cyclical dependencies.
    """

    id = "id"
    author = ProjectUser
    createdAt = "createdAt"
    text = "text"
    updatedAt = "updatedAt"


@dataclass
class Issue:
    """
    A wrapper for Issue GraphQL object.
    """

    id = "id"
    assignee = ProjectUser
    author = ProjectUser
    comments = CommentsWithoutCommentsOf
    createdAt = "createdAt"
    hasBeenSeen = "hasBeenSeen"
    issueNumber = "issueNumber"
    project = ProjectWithoutDataset
    status = "status"
    type = "type"
    updatedAt = "updatedAt"


@dataclass
class Asset:
    """
    A wrapper for Asset GraphQL object.
    """

    id = "id"
    consensusMark = "consensusMark"
    consensusMarkCompute = "consensusMarkCompute"
    consensusMarkPerCategory = "consensusMarkPerCategory"
    content = "content"
    contentJson = "contentJson"
    contentJsonCompute = "contentJsonCompute"
    createdAt = "createdAt"
    duration = "duration"
    durationCompute = "durationCompute"
    externalId = "externalId"
    honeypotMark = "honeypotMark"
    honeypotMarkCompute = "honeypotMarkCompute"
    inferenceMark = "inferenceMark"
    inferenceMarkCompute = "inferenceMarkCompute"
    isHoneypot = "isHoneypot"
    isToBeLabeledBy = "isToBeLabeledBy"
    issues = Issue
    isUsedForConsensus = "isUsedForConsensus"
    jsonContent = "jsonContent"
    jsonMetadata = "jsonMetadata"
    labels = LabelWithoutLabelOf
    latestLabel = LabelWithoutLabelOf
    latestLabelCompute = LabelWithoutLabelOf
    locks = Lock
    metadataCompute = "metadataCompute"
    metadata = "metadata"
    numberOfValidLocks = "numberOfValidLocks"
    numberOfValidLocksCompute = "numberOfValidLocksCompute"
    ocrMetadata = "ocrMetadata"
    priority = "priority"
    project = ProjectWithoutDataset
    projectId = "projectId"
    projectIdCompute = "projectIdCompute"
    readPermissionsFromLabels = "readPermissionsFromLabels"
    skippedCompute = "skippedCompute"
    skipped = "skipped"
    status = "status"
    statusCompute = "statusCompute"
    thumbnail = "thumbnail"
    thumbnailCompute = "thumbnailCompute"
    toBeLabeledBy = ProjectUser
    updatedAt = "updatedAt"


@dataclass
class Label(LabelWithoutLabelOf):
    """
    A wrapper for Label GraphQL object.
    """

    labelOf = Asset


@dataclass
class Project(ProjectWithoutDataset):
    """
    A wrapper for Project GraphQL object.
    """

    dataset = Asset


@dataclass
class Notification:
    """
    A wrapper for Notification GraphQL object.
    """

    id = "id"
    createdAt = "createdAt"
    hasBeenSeen = "hasBeenSeen"
    message = "message"
    status = "status"
    url = "url"
    userID = "userID"


@dataclass
class ProjectVersion:
    """
    A wrapper for ProjectVersion GraphQL object.
    """

    id = "id"
    createdAt = "createdAt"
    content = "content"
    name = "name"
    project = Project
    projectId = "projectId"


# pylint: enable=invalid-name



"""Organization mutations."""

import json
from typing import Optional

from typeguard import typechecked

from ...helpers import Compatible, format_result
from .queries import GQL_CREATE_ORGANIZATION, GQL_UPDATE_PROPERTIES_IN_ORGANIZATION


class MutationsOrganization:
    """Set of Organization mutations."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initializes the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    @Compatible(["v1", "v2"])
    @typechecked
    def create_organization(self, name: str, address: str, zip_code: str, city: str, country: str):
        """Create an organization.

        Each user must be linked to an organization

        Args:
            name :
            address :
            zip_code :
            city :
            country :

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        variables = {
            "data": {
                "name": name,
                "address": address,
                "zipCode": zip_code,
                "city": city,
                "country": country,
            }
        }
        result = self.auth.client.execute(GQL_CREATE_ORGANIZATION, variables)
        return format_result("data", result)

    @Compatible(["v1", "v2"])
    @typechecked
    def update_properties_in_organization(
        self,
        organization_id: str,
        name: Optional[str] = None,
        address: Optional[str] = None,
        zip_code: Optional[str] = None,
        city: Optional[str] = None,
        country: Optional[str] = None,
        license: Optional[dict] = None,
    ):  # pylint: disable=redefined-builtin
        """Modify an organization.

        Args:
            organization_id :
            name :
            address :
            license :
            zip_code :
            city :
            country :

        Returns:
            A result object which indicates if the mutation was successful,
                or an error message.
        """
        license_str = None if not license else json.dumps(license)
        variables = {"id": organization_id}
        if name is not None:
            variables["name"] = name
        if address is not None:
            variables["address"] = address
        if license_str is not None:
            variables["license"] = license_str
        if zip_code is not None:
            variables["zipCode"] = zip_code
        if city is not None:
            variables["city"] = city
        if country is not None:
            variables["country"] = country
        result = self.auth.client.execute(GQL_UPDATE_PROPERTIES_IN_ORGANIZATION, variables)
        return format_result("data", result)



"""
Tests for the client to return correct ee
"""
import os
from unittest import mock

import pytest

from kili.client import Kili
from kili.exceptions import AuthenticationFailed


class TestClient:
    """
    test the get_project function
    """

    @mock.patch.dict(os.environ, {}, clear=True)
    def test_no_api_key(self, monkeypatch):
        with pytest.raises(AuthenticationFailed):
            _ = Kili()



"""
This script permits to setup the python package.
"""

from setuptools import find_packages, setup

from kili import __version__

setup(
    # name on pypi
    name="kili",
    # code version
    version=__version__,
    # List packages
    packages=find_packages(),
    author="Kili Technology",
    author_email="contact@kili-technology.com",
    description="Python client for Kili Technology labeling tool",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "pandas",
        "click",
        "requests",
        "six",
        "tabulate",
        "tqdm",
        "typeguard",
        "pyparsing",
        "websocket-client",
    ],
    # Taking into account MANIFEST.in
    include_package_data=True,
    # Official page
    url="https://github.com/kili-technology/kili-python-sdk",
    # Some meta data:
    # https://pypi.python.org/pypi?%3Aaction=list_classifiers.
    classifiers=[
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        ("Programming Language :: Python :: " "Implementation :: CPython"),
        ("Programming Language :: Python :: " "Implementation :: PyPy"),
    ],
    entry_points={
        "console_scripts": ["kili=kili.cli:main"],
    },
)



"""Deprecation tests"""

import inspect

from kili import __version__, client


def test_catch_deprecated_elements_to_remove():
    """Raise an error if a function has a deprecation tag of a version
    that is bigger that the version that is released."""
    functions = inspect.getmembers(client.Kili, inspect.isfunction)
    functions_warning = []
    for function_name, function in functions:
        if hasattr(function, "removed_in"):
            deprecation_version = getattr(function, "removed_in")
            if __version__.split(".")[:2] >= deprecation_version.split("."):
                functions_warning.append(function_name)
    assert len(functions_warning) == 0, (
        f"Functions: {[function_name for function_name in functions_warning]}"
        "are deprecated or have a deprecated argument that should be removed"
        f"in version {__version__}"
    )



"""CLI's common helpers functions"""

import csv
import warnings
from typing import Any, Dict, List, Optional

# pylint: disable=consider-using-f-string


def dict_type_check(dict_: Dict[str, Any], type_check):
    """check if elements in row have correct type and return [row]"""
    warnings_message = ""
    for key, value in dict_.items():
        warnings_message += type_check(key, value)
    if len(warnings_message) == 0:
        return [dict_]

    warnings.warn(warnings_message + "{} will not be added.".format(list(dict_.values())[0]))
    return []


def collect_from_csv(
    csv_path: str,
    required_columns: List[str],
    optional_columns: Optional[List[str]],
    type_check_function,
):
    """read a csv to collect required_columns and optional_columns"""
    out = []
    with open(csv_path, "r", encoding="utf-8") as csv_file:
        csvreader = csv.DictReader(csv_file)
        headers = csvreader.fieldnames
        missing_columns = list(set(required_columns) - set(headers))
        if len(missing_columns) > 0:
            raise ValueError(f"{missing_columns} must be headers of the csv file: {csv_path}")
        for row in csvreader:
            out += dict_type_check(
                dict_={k: v for k, v in row.items() if k in required_columns + optional_columns},
                type_check=type_check_function,
            )

    return out


def get_external_id_from_file_path(path: str):
    """Return external_id from file's path
    ex: 'tree/leaf/file_name.txt- -> file_name
    """
    file_path = path.split("/")[-1]
    return ".".join(file_path.split(".")[:-1])


def check_exclusive_options(csv_path: Optional[str], files: Optional[List[str]]):
    """Forbid mutual use of options and argument(s)"""

    if (csv_path is not None) + (len(files) > 0) > 1:
        raise ValueError("files arguments and option --from-csv are exclusive.")

    if (csv_path is not None) + (len(files) > 0) == 0:
        raise ValueError("You must either provide file arguments or use the option --from-csv")



"""Label queries."""

import warnings
from typing import Generator, List, Optional, Union

import pandas as pd
from typeguard import typechecked

from ...constants import NO_ACCESS_RIGHT
from ...helpers import (
    Compatible,
    deprecate,
    format_result,
    fragment_builder,
    validate_category_search_query,
)
from ...orm import Label
from ...types import Label as LabelType
from ...utils.pagination import row_generator_from_paginated_calls
from ..asset import QueriesAsset
from ..project import QueriesProject
from .queries import GQL_LABELS_COUNT, gql_labels


class QueriesLabel:
    """Set of Label queries."""

    # pylint: disable=too-many-arguments,too-many-locals

    def __init__(self, auth):
        """Initialize the subclass.

        Args:
            auth: KiliAuth object
        """
        self.auth = auth

    # pylint: disable=dangerous-default-value
    @Compatible(["v1", "v2"])
    @typechecked
    @deprecate(removed_in="2.118")
    def labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        fields: List[str] = [
            "author.email",
            "author.id",
            "id",
            "jsonResponse",
            "labelType",
            "secondsToLabel",
            "skipped",
        ],
        first: Optional[int] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        id_contains: Optional[List[str]] = None,
        label_id: Optional[str] = None,
        skip: int = 0,
        skipped: Optional[bool] = None,
        type_in: Optional[List[str]] = None,
        user_id: Optional[str] = None,
        disable_tqdm: bool = False,
        as_generator: bool = False,
        category_search: Optional[str] = None,
    ) -> Union[List[dict], Generator[dict, None, None]]:
        # pylint: disable=line-too-long
        """Get a label list or a label generator from a project based on a set of criteria.

        Args:
            project_id: Identifier of the project.
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have a label whose status belongs to that list, if given.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
                Notice that the field "skipped" is deprecated since: 20/07/2022.
                It will be removed after: 31/08/2022.
            first: Maximum number of labels to return.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            id_contains: Filters out labels not belonging to that list. If empty, no filtering is applied.
            label_id: Identifier of the label.
            skip: Number of labels to skip (they are ordered by their date of creation, first to last).
            skipped: Returned labels should have a label which is skipped
                Notice that the field "skipped" is deprecated since: 20/07/2022.
                It will be removed after: 31/08/2022.
            type_in: Returned labels should have a label whose type belongs to that list, if given.
            user_id: Identifier of the user.
            disable_tqdm: If `True`, the progress bar will be disabled
            as_generator: If `True`, a generator on the labels is returned.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            A result object which contains the query if it was successful, else an error message.

        Examples:
            >>> kili.labels(project_id=project_id, fields=['jsonResponse', 'labelOf.externalId']) # returns a list of all labels of a project and their assets external ID
            >>> kili.labels(project_id=project_id, fields=['jsonResponse'], as_generator=True) # returns a generator of all labels of a project

        !!! example "How to filter based on label categories"
            The search query is composed of logical expressions following this format:

                [job_name].[category_name].count [comparaison_operator] [value]
            where:

            - `[job_name]` is the name of the job in the interface
            - `[category_name]` is the name of the category in the interface for this job
            - `[comparaison_operator]` can be one of: [`==`, `>=`, `<=`, `<`, `>`]
            - `[value]` is an integer that represents the count of such objects of the given category in the label

            These operations can be separated by OR and AND operators

            Example:

                category_search = `JOB_CLASSIF.CATEGORY_A.count > 0`
                category_search = `JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0`
                category_search = `(JOB_CLASSIF.CATEGORY_A.count > 0 OR JOB_NER.CATEGORY_B.count > 0) AND JOB_BBOX.CATEGORY_C.count > 10`
        """
        if "skipped" in fields or skipped is not None:
            message = """
                The field "skipped" is deprecated since: 20/07/2022.
                It will be removed after: 31/08/2022.
                """
            warnings.warn(message, DeprecationWarning)

        saved_args = locals()
        count_args = {
            k: v
            for (k, v) in saved_args.items()
            if k
            not in [
                "as_generator",
                "disable_tqdm",
                "fields",
                "first",
                "id_contains",
                "self",
                "skip",
                "message",
            ]
        }

        # using tqdm with a generator is messy, so it is always disabled
        disable_tqdm = disable_tqdm or as_generator

        if category_search:
            validate_category_search_query(category_search)

        payload_query = {
            "where": {
                "id": label_id,
                "asset": {
                    "id": asset_id,
                    "externalIdIn": asset_external_id_in,
                    "statusIn": asset_status_in,
                },
                "project": {
                    "id": project_id,
                },
                "user": {
                    "id": user_id,
                },
                "createdAt": created_at,
                "createdAtGte": created_at_gte,
                "createdAtLte": created_at_lte,
                "authorIn": author_in,
                "honeypotMarkGte": honeypot_mark_gte,
                "honeypotMarkLte": honeypot_mark_lte,
                "idIn": id_contains,
                "search": category_search,
                "skipped": skipped,
                "typeIn": type_in,
            },
        }

        labels_generator = row_generator_from_paginated_calls(
            skip,
            first,
            self.count_labels,
            count_args,
            self._query_labels,
            payload_query,
            fields,
            disable_tqdm,
        )

        if as_generator:
            return labels_generator
        return list(labels_generator)

    def _query_labels(self, skip: int, first: int, payload: dict, fields: List[str]):

        payload.update({"skip": skip, "first": first})
        _gql_labels = gql_labels(fragment_builder(fields, LabelType))
        result = self.auth.client.execute(_gql_labels, payload)
        return format_result("data", result, Label)

    # pylint: disable=dangerous-default-value
    @typechecked
    @deprecate(removed_in="2.118")
    def export_labels_as_df(
        self,
        project_id: str,
        fields: List[str] = [
            "author.email",
            "author.id",
            "createdAt",
            "id",
            "labelType",
            "skipped",
        ],
        asset_fields: List[str] = ["externalId"],
    ) -> pd.DataFrame:
        # pylint: disable=line-too-long
        """Get the labels of a project as a pandas DataFrame.

        Args:
            project_id: Identifier of the project
            fields: All the fields to request among the possible fields for the labels.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#label) for all possible fields.
                Notice that the field "skipped" is deprecated since: 20/07/2022.
                It will be removed after: 31/08/2022.
            asset_fields: All the fields to request among the possible fields for the assets.
                See [the documentation](https://docs.kili-technology.com/reference/graphql-api#asset) for all possible fields.

        Returns:
            pandas DataFrame containing the labels.
        """
        if "skipped" in fields:
            message = """
                The field "skipped" is deprecated since: 20/07/2022.
                It will be removed after: 31/08/2022.
                """
            warnings.warn(message, DeprecationWarning)

        projects = QueriesProject(self.auth).projects(project_id)
        assert len(projects) == 1, NO_ACCESS_RIGHT
        assets = QueriesAsset(self.auth).assets(
            project_id=project_id,
            fields=asset_fields + ["labels." + field for field in fields],
        )
        labels = [
            dict(
                label,
                **dict((f"asset_{key}", asset[key]) for key in asset if key != "labels"),
            )
            for asset in assets
            for label in asset["labels"]
        ]
        labels_df = pd.DataFrame(labels)
        return labels_df

    @Compatible(["v1", "v2"])
    @typechecked
    @deprecate(removed_in="2.118")
    def count_labels(
        self,
        project_id: str,
        asset_id: Optional[str] = None,
        asset_status_in: Optional[List[str]] = None,
        asset_external_id_in: Optional[List[str]] = None,
        author_in: Optional[List[str]] = None,
        created_at: Optional[str] = None,
        created_at_gte: Optional[str] = None,
        created_at_lte: Optional[str] = None,
        honeypot_mark_gte: Optional[float] = None,
        honeypot_mark_lte: Optional[float] = None,
        label_id: Optional[str] = None,
        skipped: Optional[bool] = None,
        type_in: Optional[List[str]] = None,
        user_id: Optional[str] = None,
        category_search: Optional[str] = None,
    ) -> int:
        # pylint: disable=line-too-long
        """Get the number of labels for the given parameters.

        Args:
            asset_id: Identifier of the asset.
            asset_status_in: Returned labels should have a status that belongs to that list, if given.
                Possible choices : `TODO`, `ONGOING`, `LABELED` or `REVIEWED`
            asset_external_id_in: Returned labels should have an external id that belongs to that list, if given.
            author_in: Returned labels should have a label whose status belongs to that list, if given.
            created_at: Returned labels should have a label whose creation date is equal to this date.
            created_at_gte: Returned labels should have a label whose creation date is greater than this date.
            created_at_lte: Returned labels should have a label whose creation date is lower than this date.
            honeypot_mark_gte: Returned labels should have a label whose honeypot is greater than this number.
            honeypot_mark_lte: Returned labels should have a label whose honeypot is lower than this number.
            label_id: Identifier of the label.
            project_id: Identifier of the project.
            skipped: Returned labels should have a label which is skipped
                    Notice that the field "skipped" is deprecated since: 20/07/2022.
                    It will be removed after: 31/08/2022.
            type_in: Returned labels should have a label whose type belongs to that list, if given.
            user_id: Identifier of the user.

        !!! info "Dates format"
            Date strings should have format: "YYYY-MM-DD"

        Returns:
            The number of labels with the parameters provided
        """
        if skipped is not None:
            message = """
                The field "skipped" is deprecated since: 20/07/2022.
                It will be removed after: 31/08/2022.
                """
            warnings.warn(message, DeprecationWarning)

        if category_search:
            validate_category_search_query(category_search)

        variables = {
            "where": {
                "id": label_id,
                "asset": {
                    "id": asset_id,
                    "externalIdIn": asset_external_id_in,
                    "statusIn": asset_status_in,
                },
                "project": {
                    "id": project_id,
                },
                "user": {
                    "id": user_id,
                },
                "createdAt": created_at,
                "createdAtGte": created_at_gte,
                "createdAtLte": created_at_lte,
                "authorIn": author_in,
                "honeypotMarkGte": honeypot_mark_gte,
                "honeypotMarkLte": honeypot_mark_lte,
                "search": category_search,
                "skipped": skipped,
                "typeIn": type_in,
            }
        }
        result = self.auth.client.execute(GQL_LABELS_COUNT, variables)
        count = format_result("data", result)
        return count
